import importlib.util
import platform
import sys

import keras

from keras_rs.src.api_export import keras_rs_export

# JAX distributed embedding is only available on linux_x86_64, and only if
# jax-tpu-embedding is installed.
jax_tpu_embedding_spec = importlib.util.find_spec("jax_tpu_embedding")
if (
    keras.backend.backend() == "jax"
    and sys.platform == "linux"
    and platform.machine().lower() == "x86_64"
    and jax_tpu_embedding_spec is not None
):
    from keras_rs.src.layers.embedding.jax.distributed_embedding import (
        DistributedEmbedding as BackendDistributedEmbedding,
    )
elif keras.backend.backend() == "tensorflow":
    from keras_rs.src.layers.embedding.tensorflow.distributed_embedding import (
        DistributedEmbedding as BackendDistributedEmbedding,
    )
else:
    from keras_rs.src.layers.embedding.base_distributed_embedding import (
        DistributedEmbedding as BackendDistributedEmbedding,
    )


@keras_rs_export("keras_rs.layers.DistributedEmbedding")
class DistributedEmbedding(BackendDistributedEmbedding):
    pass



from typing import Any, Callable, Sequence, TypeAlias

import keras
import tensorflow as tf

from keras_rs.src import types
from keras_rs.src.layers.embedding import base_distributed_embedding
from keras_rs.src.layers.embedding import distributed_embedding_config
from keras_rs.src.layers.embedding.tensorflow import config_conversion
from keras_rs.src.utils import keras_utils

FeatureConfig = distributed_embedding_config.FeatureConfig
TableConfig = distributed_embedding_config.TableConfig

# Placeholder of tf.tpu.experimental.embedding._Optimizer which is not exposed.
TfTpuOptimizer: TypeAlias = Any


GRADIENT_TRAP_DUMMY_NAME = "_gradient_trap_dummy"

EMBEDDING_FEATURE_V1 = tf.tpu.experimental.HardwareFeature.EmbeddingFeature.V1
EMBEDDING_FEATURE_V2 = tf.tpu.experimental.HardwareFeature.EmbeddingFeature.V2
UNSUPPORTED = tf.tpu.experimental.HardwareFeature.EmbeddingFeature.UNSUPPORTED


class DistributedEmbedding(base_distributed_embedding.DistributedEmbedding):
    """TensorFlow implementation of the TPU embedding layer."""

    def __init__(
        self,
        feature_configs: types.Nested[
            FeatureConfig | tf.tpu.experimental.embedding.FeatureConfig
        ],
        *,
        table_stacking: (
            str | Sequence[str] | Sequence[Sequence[str]]
        ) = "auto",
        **kwargs: Any,
    ) -> None:
        # Intercept arguments that are supported only on TensorFlow.
        self._optimizer = kwargs.pop("optimizer", None)
        self._pipeline_execution_with_tensor_core = kwargs.pop(
            "pipeline_execution_with_tensor_core", False
        )
        self._sparse_core_embedding_config = kwargs.pop(
            "sparse_core_embedding_config", None
        )

        # Mark as True by default for `_verify_input_shapes`. This will be
        # updated in `_sparsecore_init` if applicable.
        self._using_keras_rs_configuration = True

        super().__init__(
            feature_configs, table_stacking=table_stacking, **kwargs
        )

    def _is_tpu_strategy(self, strategy: tf.distribute.Strategy) -> bool:
        return isinstance(
            strategy,
            (tf.distribute.TPUStrategy, tf.distribute.experimental.TPUStrategy),
        )

    def _has_sparsecore(self) -> bool:
        strategy = tf.distribute.get_strategy()
        if self._is_tpu_strategy(strategy):
            tpu_embedding_feature = (
                strategy.extended.tpu_hardware_feature.embedding_feature
            )
            return tpu_embedding_feature in (
                EMBEDDING_FEATURE_V2,
                EMBEDDING_FEATURE_V1,
            )
        return False

    @keras_utils.no_automatic_dependency_tracking
    def _sparsecore_init(
        self,
        feature_configs: dict[
            str,
            FeatureConfig | tf.tpu.experimental.embedding.FeatureConfig,
        ],
        table_stacking: str | Sequence[str] | Sequence[Sequence[str]],
    ) -> None:
        self._table_stacking = table_stacking

        strategy = tf.distribute.get_strategy()
        if not self._is_tpu_strategy(strategy):
            raise ValueError(
                "Placement to sparsecore was requested, however, we are not "
                "running under a TPU strategy."
            )

        tpu_embedding_feature = (
            strategy.extended.tpu_hardware_feature.embedding_feature
        )

        self._using_keras_rs_configuration = isinstance(
            next(iter(feature_configs.values())), FeatureConfig
        )
        if self._using_keras_rs_configuration:
            if self._sparse_core_embedding_config is not None:
                raise ValueError(
                    "The `sparse_core_embedding_config` argument is only "
                    "supported when using "
                    "`tf.tpu.experimental.embedding.FeatureConfig` instances "
                    "for the configuration."
                )
            self._tpu_feature_configs, self._sparse_core_embedding_config = (
                config_conversion.keras_to_tf_tpu_configuration(
                    feature_configs,
                    table_stacking,
                    strategy.num_replicas_in_sync,
                )
            )
            if tpu_embedding_feature == EMBEDDING_FEATURE_V1:
                # Remove auto-generated SparseCoreEmbeddingConfig, which is not
                # used.
                self._sparse_core_embedding_config = None
        else:
            if table_stacking != "auto":
                raise ValueError(
                    "The `table_stacking` argument is not supported when using "
                    "`tf.tpu.experimental.embedding.FeatureConfig` for the "
                    "configuration. You can use the `disable_table_stacking` "
                    "attribute of "
                    "`tf.tpu.experimental.embedding.SparseCoreEmbeddingConfig` "
                    "to disable table stacking."
                )
            if (
                tpu_embedding_feature == EMBEDDING_FEATURE_V1
                and self._sparse_core_embedding_config is not None
            ):
                raise ValueError(
                    "The `sparse_core_embedding_config` argument is not "
                    "supported with this TPU generation."
                )
            self._tpu_feature_configs = (
                config_conversion.clone_tf_tpu_feature_configs(feature_configs)
            )

        self._tpu_optimizer = config_conversion.to_tf_tpu_optimizer(
            self._optimizer
        )

        if tpu_embedding_feature == EMBEDDING_FEATURE_V1:
            self._tpu_embedding = tf.tpu.experimental.embedding.TPUEmbedding(
                self._tpu_feature_configs,
                self._tpu_optimizer,
                self._pipeline_execution_with_tensor_core,
            )
            self._v1_call_id = 0
        elif tpu_embedding_feature == EMBEDDING_FEATURE_V2:
            self._tpu_embedding = tf.tpu.experimental.embedding.TPUEmbeddingV2(
                self._tpu_feature_configs,
                self._tpu_optimizer,
                self._pipeline_execution_with_tensor_core,
                self._sparse_core_embedding_config,
            )
        elif tpu_embedding_feature == UNSUPPORTED:
            raise ValueError(
                "Placement to sparsecore was requested, however, this TPU does "
                "not support it."
            )
        elif tpu_embedding_feature != UNSUPPORTED:
            raise ValueError(
                f"Unsupported TPU embedding feature: {tpu_embedding_feature}."
            )

        # We need at least one trainable variable for the gradient trap to work.
        # Note that the Python attribute name "_gradient_trap_dummy" should
        # match the name of the variable GRADIENT_TRAP_DUMMY_NAME.
        self._gradient_trap_dummy = self.add_weight(
            name=GRADIENT_TRAP_DUMMY_NAME,
            shape=(1,),
            initializer=tf.zeros_initializer(),
            trainable=True,
            dtype=tf.float32,
        )

    def compute_output_shape(
        self, input_shapes: types.Nested[types.Shape]
    ) -> types.Nested[types.Shape]:
        if self._using_keras_rs_configuration:
            return super().compute_output_shape(input_shapes)

        def _compute_output_shape(
            feature_config: tf.tpu.experimental.embedding.FeatureConfig,
            input_shape: types.Shape,
        ) -> types.Shape:
            if len(input_shape) < 1:
                raise ValueError(
                    f"Received input shape {input_shape}. Rank must be 1 or "
                    "above."
                )
            max_sequence_length: int = feature_config.max_sequence_length
            embed_dim = feature_config.table.dim
            if (
                feature_config.output_shape is not None
                and feature_config.output_shape.rank is not None
            ):
                return tuple(feature_config.output_shape.as_list())
            elif (
                len(input_shape) == 2
                and input_shape[-1] != 1
                and max_sequence_length > 0
            ):
                # Update the input shape with the max sequence length. Only
                # update when:
                # 1. Input feature is 2D ragged or sparse tensor.
                # 2. Output shape is not set and max sequence length is set.
                return tuple(input_shape[:-1]) + (
                    max_sequence_length,
                    embed_dim,
                )
            elif len(input_shape) == 1:
                return tuple(input_shape) + (embed_dim,)
            else:
                return tuple(input_shape[:-1]) + (embed_dim,)

        output_shapes: types.Nested[types.Shape] = (
            keras.tree.map_structure_up_to(
                self._feature_configs,
                _compute_output_shape,
                self._feature_configs,
                input_shapes,
            )
        )
        return output_shapes

    def _sparsecore_build(self, input_shapes: dict[str, types.Shape]) -> None:
        if isinstance(
            self._tpu_embedding, tf.tpu.experimental.embedding.TPUEmbedding
        ):
            tf_input_shapes = keras.tree.map_shape_structure(
                tf.TensorShape, input_shapes
            )
            tpu_embedding_build = tf.autograph.to_graph(
                self._tpu_embedding.build, recursive=False
            )
            tpu_embedding_build(
                self._tpu_embedding, per_replica_input_shapes=tf_input_shapes
            )
        elif isinstance(
            self._tpu_embedding, tf.tpu.experimental.embedding.TPUEmbeddingV2
        ):
            self._tpu_embedding.build()

    def _sparsecore_call(
        self,
        inputs: dict[str, types.Tensor],
        weights: dict[str, types.Tensor] | None = None,
        training: bool = False,
    ) -> dict[str, types.Tensor]:
        del training  # Unused.
        strategy = tf.distribute.get_strategy()
        if not self._is_tpu_strategy(strategy):
            raise RuntimeError(
                "DistributedEmbedding needs to be called under a TPUStrategy "
                "for features placed on the embedding feature but is being "
                f"called under strategy {strategy}. Please use `strategy.run` "
                "when calling this layer."
            )
        if isinstance(
            self._tpu_embedding, tf.tpu.experimental.embedding.TPUEmbedding
        ):
            return self._tpu_embedding_lookup_v1(
                self._tpu_embedding, inputs, weights
            )
        elif isinstance(
            self._tpu_embedding, tf.tpu.experimental.embedding.TPUEmbeddingV2
        ):
            return self._tpu_embedding_lookup_v2(
                self._tpu_embedding, inputs, weights
            )
        else:
            raise ValueError(
                "DistributedEmbedding is receiving features to lookup on the "
                "TPU embedding feature but no such feature was configured."
            )

    def _sparsecore_get_embedding_tables(self) -> dict[str, types.Tensor]:
        tables: dict[str, types.Tensor] = {}
        strategy = tf.distribute.get_strategy()
        if not self._is_tpu_strategy(strategy):
            raise RuntimeError(
                "`DistributedEmbedding.get_embedding_tables` needs to be "
                "called under the TPUStrategy that DistributedEmbedding was "
                f"created with, but is being called under strategy {strategy}. "
                "Please use `with strategy.scope()` when calling "
                "`get_embedding_tables`."
            )

        tpu_hardware = strategy.extended.tpu_hardware_feature
        num_sc_per_device = tpu_hardware.num_embedding_devices_per_chip
        num_shards = strategy.num_replicas_in_sync * num_sc_per_device

        def populate_table(
            feature_config: tf.tpu.experimental.embedding.FeatureConfig,
        ) -> None:
            table_name = feature_config.table.name
            if table_name in tables:
                return

            embedding_dim = feature_config.table.dim
            table = self._tpu_embedding.embedding_tables[table_name]

            # This table has num_sparse_cores mod shards, so we need to slice,
            # reconcat and reshape.
            table_shards = [
                shard.numpy()[:, :embedding_dim] for shard in table.values
            ]
            full_table = keras.ops.concatenate(table_shards, axis=0)
            full_table = keras.ops.concatenate(
                keras.ops.split(full_table, num_shards, axis=0), axis=1
            )
            full_table = keras.ops.reshape(full_table, [-1, embedding_dim])
            tables[table_name] = full_table[
                : feature_config.table.vocabulary_size, :
            ]

        keras.tree.map_structure(populate_table, self._tpu_feature_configs)
        return tables

    def _verify_input_shapes(
        self, input_shapes: types.Nested[types.Shape]
    ) -> None:
        if self._using_keras_rs_configuration:
            return super()._verify_input_shapes(input_shapes)
        # `tf.tpu.experimental.embedding.FeatureConfig` does not provide any
        # information about the input shape, so there is nothing to verify.

    def _tpu_embedding_lookup_v1(
        self,
        tpu_embedding: tf.tpu.experimental.embedding.TPUEmbedding,
        inputs: dict[str, types.Tensor],
        weights: dict[str, types.Tensor] | None = None,
    ) -> dict[str, types.Tensor]:
        # Each call to this function increments the _v1_call_id by 1, this
        # allows us to tag each of the main embedding ops with this call id so
        # that we know during graph rewriting passes which ops correspond to the
        # same layer call.
        self._v1_call_id += 1
        name = str(self._v1_call_id)

        # Set training to true, even during eval. When name is set, this will
        # trigger a pass that updates the training based on if there is a send
        # gradients with the same name.
        tpu_embedding.enqueue(inputs, weights, training=True, name=name)

        @tf.custom_gradient  # type: ignore
        def gradient_trap(
            dummy: types.Tensor,
        ) -> tuple[
            list[types.Tensor], Callable[[tuple[types.Tensor]], types.Tensor]
        ]:
            """Register a gradient function for activation."""
            activations = tpu_embedding.dequeue(name=name)

            def grad(*grad_wrt_activations: types.Tensor) -> types.Tensor:
                """Gradient function."""
                # Since the output were flattened, the gradients are also
                # flattened. Pack them back into the correct nested structure.
                gradients = tf.nest.pack_sequence_as(
                    self._placement_to_path_to_feature_config["sparsecore"],
                    grad_wrt_activations,
                )
                tpu_embedding.apply_gradients(gradients, name=name)

                # This is the gradient for the input variable.
                return tf.zeros_like(dummy)

            # Custom gradient functions don't like nested structures of tensors,
            # so we flatten them here.
            return tf.nest.flatten(activations), grad

        activations_with_trap = gradient_trap(self._gradient_trap_dummy.value)
        result: dict[str, types.Tensor] = tf.nest.pack_sequence_as(
            self._placement_to_path_to_feature_config["sparsecore"],
            activations_with_trap,
        )
        return result

    def _tpu_embedding_lookup_v2(
        self,
        tpu_embedding: tf.tpu.experimental.embedding.TPUEmbeddingV2,
        inputs: dict[str, types.Tensor],
        weights: dict[str, types.Tensor] | None = None,
    ) -> dict[str, types.Tensor]:
        @tf.custom_gradient  # type: ignore
        def gradient_trap(
            dummy: types.Tensor,
        ) -> tuple[
            list[types.Tensor], Callable[[tuple[types.Tensor]], types.Tensor]
        ]:
            """Register a gradient function for activation."""
            activations, preserved_result = tpu_embedding(inputs, weights)

            def grad(*grad_wrt_activations: types.Tensor) -> types.Tensor:
                """Gradient function."""
                # Since the output were flattened, the gradients are also
                # flattened. Pack them back into the correct nested structure.
                gradients = tf.nest.pack_sequence_as(
                    self._placement_to_path_to_feature_config["sparsecore"],
                    grad_wrt_activations,
                )
                tpu_embedding.apply_gradients(
                    gradients, preserved_outputs=preserved_result
                )
                # This is the gradient for the input variable.
                return tf.zeros_like(dummy)

            # Custom gradient functions don't like nested structures of tensors,
            # so we flatten them here.
            return tf.nest.flatten(activations), grad

        activations_with_trap = gradient_trap(self._gradient_trap_dummy)
        result: dict[str, types.Tensor] = tf.nest.pack_sequence_as(
            self._placement_to_path_to_feature_config["sparsecore"],
            activations_with_trap,
        )
        return result

    def _trackable_children(
        self, save_type: str = "checkpoint", **kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        # Remove dummy variable, we don't want it in checkpoints.
        children: dict[str, Any] = super()._trackable_children(
            save_type, **kwargs
        )
        children.pop(GRADIENT_TRAP_DUMMY_NAME, None)
        return children


DistributedEmbedding.__doc__ = (
    base_distributed_embedding.DistributedEmbedding.__doc__
)



import collections
import dataclasses
import importlib.util
import typing
from typing import Any, Sequence

import keras
import numpy as np
from keras.src import backend

from keras_rs.src import types
from keras_rs.src.layers.embedding import distributed_embedding_config
from keras_rs.src.layers.embedding import embed_reduce
from keras_rs.src.utils import keras_utils

FeatureConfig = distributed_embedding_config.FeatureConfig
TableConfig = distributed_embedding_config.TableConfig
EmbedReduce = embed_reduce.EmbedReduce


SUPPORTED_PLACEMENTS = ("auto", "default_device", "sparsecore")


@dataclasses.dataclass(eq=True, unsafe_hash=True, order=True)
class PlacementAndPath:
    placement: str
    path: str


def _ragged_to_dense_inputs(
    inputs: Any, weights: Any | None = None, dense_row_length: int | None = None
) -> Any:
    """Converts a ragged set of inputs and weights to dense.

    If inputs are ragged and weights are `None`, will create a dense set of
    weights to mask out the new padded values.

    If inputs are not ragged, returns the original `inputs` and `weights`
    unmodified.

    Args:
        inputs: The inputs array.
        weights: The optional weights array.
        dense_row_length: The output dense row length.  If None, uses the length
             of the longest row of the input.

    Returns:
        Tuple of new (inputs, weights).  If the input is a ragged array, returns
        dense numpy arrays. Otherwise, returns the original input and weights.
    """
    x = inputs
    w = weights
    # tf.Ragged or other .numpy()-able types.
    if hasattr(x, "numpy") and callable(getattr(x, "numpy")):
        x = x.numpy()

    # Ragged numpy array to dense numpy array.
    if isinstance(x, np.ndarray) and len(x) > 0 and x.dtype == np.ndarray:
        # Maybe convert weights to numpy.
        if (
            w is not None
            and hasattr(w, "numpy")
            and callable(getattr(w, "numpy"))
        ):
            w = w.numpy()

        if dense_row_length is None:
            # Use length of longest row.
            dense_row_length = max([len(row) for row in x])

        output = np.zeros((len(x), dense_row_length), dtype=x[0].dtype)
        for i, row in enumerate(x):
            output[i, : len(row)] = row

        output_weights = np.zeros((len(x), dense_row_length), dtype=np.float32)
        if w is None:
            for i, row in enumerate(x):
                output_weights[i, : len(row)] = 1.0
        else:
            for i, row in enumerate(w):
                output_weights[i, : len(row)] = row

        return output, output_weights

    # Convert symbolic ragged/sparse keras tensors to dense tensors.
    if isinstance(x, keras.KerasTensor) and (x.ragged or x.sparse):
        inputs = keras.ops.convert_to_tensor(x, ragged=False)
        weights = keras.ops.convert_to_tensor(x, dtype="float32", ragged=False)

    # If not a ragged array, return the original, unmodified.
    return inputs, weights


class DistributedEmbedding(keras.layers.Layer):
    """DistributedEmbedding, a layer for accelerated large embedding lookups.

    ---

    ## Note: `DistributedEmbedding` is in Preview.

    ---

    `DistributedEmbedding` is a layer optimized for TPU chips with SparseCore
    and can dramatically improve the speed of embedding lookups and embedding
    training. It works by combining multiple lookups into one invocation, and by
    sharding the embedding tables across the available chips. Note that one will
    only see performance benefits for embedding tables that are large enough to
    to require sharding because they don't fit on a single chip. More details
    are provided in the "Placement" section below.

    On other hardware, GPUs, CPUs and TPUs without SparseCore,
    `DistributedEmbedding` provides the same API without any specific
    acceleration. No particular distribution scheme is applied besides the one
    set via `keras.distribution.set_distribution`.

    `DistributedEmbedding` embeds sequences of inputs and reduces them to a
    single embedding by applying a configurable combiner function.

    ### Configuration

    #### Features and tables

    A `DistributedEmbedding` embedding layer is configured via a set of
    `keras_rs.layers.FeatureConfig` objects, which themselves refer to
    `keras_rs.layers.TableConfig` objects.

    - `TableConfig` defines an embedding table with parameters such as its
      vocabulary size, embedding dimension, as well as a combiner for reduction
      and optimizer for training.
    - `FeatureConfig` defines what input features the `DistributedEmbedding`
      will handle and which embedding table to use. Note that multiple features
      can use the same embedding table.

    ```python
    table1 = keras_rs.layers.TableConfig(
        name="table1",
        vocabulary_size=TABLE1_VOCABULARY_SIZE,
        embedding_dim=TABLE1_EMBEDDING_SIZE,
        placement="auto",
    )
    table2 = keras_rs.layers.TableConfig(
        name="table2",
        vocabulary_size=TABLE2_VOCABULARY_SIZE,
        embedding_dim=TABLE2_EMBEDDING_SIZE,
        placement="auto",
    )

    feature1 = keras_rs.layers.FeatureConfig(
        name="feature1",
        table=table1,
        input_shape=(GLOBAL_BATCH_SIZE,),
        output_shape=(GLOBAL_BATCH_SIZE, TABLE1_EMBEDDING_SIZE),
    )
    feature2 = keras_rs.layers.FeatureConfig(
        name="feature2",
        table=table2,
        input_shape=(GLOBAL_BATCH_SIZE,),
        output_shape=(GLOBAL_BATCH_SIZE, TABLE2_EMBEDDING_SIZE),
    )

    feature_configs = {
        "feature1": feature1,
        "feature2": feature2,
    }

    embedding = keras_rs.layers.DistributedEmbedding(feature_configs)
    ```

    #### Optimizers

    Each embedding table within `DistributedEmbedding` uses its own optimizer
    for training, which is independent from the optimizer set on the model via
    `model.compile()`.

    Note that not all optimizers are supported. Currently, the following are
    supported on all backends and accelerators:

    - `keras.optimizers.Adagrad`
    - `keras.optimizers.Adam`
    - `keras.optimizers.Ftrl`
    - `keras.optimizers.SGD`

    Also, not all parameters of the optimizers are supported (e.g. the
    `nesterov` option of `SGD`). An error is raised when an unsupported
    optimizer or an unsupported optimizer parameter is used.

    #### Placement

    Each embedding table within `DistributedEmbedding` can be either placed on
    the SparseCore chip or the default device placement for the accelerator
    (e.g. HBM of the Tensor Cores on TPU). This is controlled by the `placement`
    attribute of `keras_rs.layers.TableConfig`.

    - A placement of `"sparsecore"` indicates that the table should be placed on
      the SparseCore chips. An error is raised if this option is selected and
      there are no SparseCore chips.
    - A placement of `"default_device"` indicates that the table should not be
      placed on SparseCore, even if available. Instead the table is placed on
      the device where the model normally goes, i.e. the HBM on TPUs and GPUs.
      In this case, if applicable, the table is distributed using the scheme set
      via `keras.distribution.set_distribution`. On GPUs, CPUs and TPUs without
      SparseCore, this is the only placement available, and is the one selected
      by `"auto"`.
    - A placement of `"auto"` indicates to use `"sparsecore"` if available, and
      `"default_device"` otherwise. This is the default when not specified.

    To optimize performance on TPU:

    - Tables that are so large that they need to be sharded should use the
      `"sparsecore"` placement.
    - Tables that are small enough should use `"default_device"` and should
      typically be replicated across TPUs by using the
      `keras.distribution.DataParallel` distribution option.

    ### Usage with TensorFlow on TPU with SpareCore

    #### Inputs

    In addition to `tf.Tensor`, `DistributedEmbedding` accepts `tf.RaggedTensor`
    and `tf.SparseTensor` as inputs for the embedding lookups. Ragged tensors
    must be ragged in the dimension with index 1. Note that if weights are
    passed, each weight tensor must be of the same class as the inputs for that
    particular feature and use the exact same ragged row lenghts for ragged
    tensors, and the same indices for sparse tensors. All the output of
    `DistributedEmbedding` are dense tensors.

    #### Setup

    To use `DistributedEmbedding` on TPUs with TensorFlow, one must use a
    `tf.distribute.TPUStrategy`. The `DistributedEmbedding` layer must be
    created under the `TPUStrategy`.

    ```python
    resolver = tf.distribute.cluster_resolver.TPUClusterResolver(tpu="local")
    topology = tf.tpu.experimental.initialize_tpu_system(resolver)
    device_assignment = tf.tpu.experimental.DeviceAssignment.build(
        topology, num_replicas=resolver.get_tpu_system_metadata().num_cores
    )
    strategy = tf.distribute.TPUStrategy(
        resolver, experimental_device_assignment=device_assignment
    )

    with strategy.scope():
        embedding = keras_rs.layers.DistributedEmbedding(feature_configs)
    ```

    #### Usage in a Keras model

    To use Keras' `model.fit()`, one must compile the model under the
    `TPUStrategy`. Then, `model.fit()`, `model.evaluate()` or `model.predict()`
    can be called directly. The Keras model takes care of running the model
    using the strategy and also automatically distributes the dataset.

    ```python
    with strategy.scope():
        embedding = keras_rs.layers.DistributedEmbedding(feature_configs)
        model = create_model(embedding)
        model.compile(loss=keras.losses.MeanSquaredError(), optimizer="adam")

    model.fit(dataset, epochs=10)
    ```

    #### Direct invocation

    `DistributedEmbedding` must be invoked via a `strategy.run` call nested in a
    `tf.function`.

    ```python
    @tf.function
    def embedding_wrapper(tf_fn_inputs, tf_fn_weights=None):
        def strategy_fn(st_fn_inputs, st_fn_weights):
            return embedding(st_fn_inputs, st_fn_weights)

        return strategy.run(strategy_fn, args=(tf_fn_inputs, tf_fn_weights)))

    embedding_wrapper(my_inputs, my_weights)
    ```

    When using a dataset, the dataset must be distributed. The iterator can then
    be passed to the `tf.function` that uses `strategy.run`.

    ```python
    dataset = strategy.experimental_distribute_dataset(dataset)

    @tf.function
    def run_loop(iterator):
        def step(data):
            (inputs, weights), labels = data
            with tf.GradientTape() as tape:
                result = embedding(inputs, weights)
                loss = keras.losses.mean_squared_error(labels, result)
            tape.gradient(loss, embedding.trainable_variables)
            return result

        for _ in tf.range(4):
            result = strategy.run(step, args=(next(iterator),))

    run_loop(iter(dataset))
    ```

    ### Usage with JAX on TPU with SpareCore

    #### Setup

    To use `DistributedEmbedding` on TPUs with JAX, one must create and set a
    Keras `Distribution`.
    ```python
    distribution = keras.distribution.DataParallel(devices=jax.device("tpu"))
    keras.distribution.set_distribution(distribution)
    ```

    #### Inputs

    For JAX, inputs can either be dense tensors, or ragged (nested) NumPy
    arrays.  To enable `jit_compile = True`, one must explicitly call
    `layer.preprocess(...)` on the inputs, and then feed the preprocessed
    output to the model.  See the next section on preprocessing for details.

    Ragged input arrays must be ragged in the dimension with index 1. Note that
    if weights are passed, each weight tensor must be of the same class as the
    inputs for that particular feature and use the exact same ragged row lengths
    for ragged tensors. All the output of `DistributedEmbedding` are dense
    tensors.

    #### Preprocessing

    In JAX, SparseCore usage requires specially formatted data that depends
    on properties of the available hardware.  This data reformatting
    currently does not support jit-compilation, so must be applied _prior_
    to passing data into a model.

    Preprocessing works on dense or ragged NumPy arrays, or on tensors that are
    convertible to dense or ragged NumPy arrays like `tf.RaggedTensor`.

    One simple way to add preprocessing is to append the function to an input
    pipeline by using a python generator.
    ```python
    # Create the embedding layer.
    embedding_layer = DistributedEmbedding(feature_configs)

    # Add preprocessing to a data input pipeline.
    def preprocessed_dataset_generator(dataset):
        for (inputs, weights), labels in iter(dataset):
            yield embedding_layer.preprocess(
                inputs, weights, training=True
            ), labels

    preprocessed_train_dataset = preprocessed_dataset_generator(train_dataset)
    ```
    This explicit preprocessing stage combines the input and optional weights,
    so the new data can be passed directly into the `inputs` argument of the
    layer or model.

    **NOTE**: When working in a multi-host setting with data parallelism, the
    data needs to be sharded properly across hosts.  If the original dataset is
    of type `tf.data.Dataset`, it will need to be manually sharded _prior_ to
    applying the preprocess generator:
    ```python
    # Manually shard the dataset across hosts.
    train_dataset = distribution.distribute_dataset(train_dataset)
    distribution.auto_shard_dataset = False  # Dataset is already sharded.

    # Add a preprocessing stage to the distributed data input pipeline.
    train_dataset = preprocessed_dataset_generator(train_dataset)
    ```
    If the original dataset is _not_ a `tf.data.Dataset`, it must already be
    pre-sharded across hosts.

    #### Usage in a Keras model

    Once the global distribution is set and the input preprocessing pipeline
    is defined, model training can proceed as normal.  For example:
    ```python
    # Construct, compile, and fit the model using the preprocessed data.
    model = keras.Sequential(
      [
        embedding_layer,
        keras.layers.Dense(2),
        keras.layers.Dense(3),
        keras.layers.Dense(4),
      ]
    )
    model.compile(optimizer="adam", loss="mse", jit_compile=True)
    model.fit(preprocessed_train_dataset, epochs=10)
    ```

    #### Direct invocation

    The `DistributedEmbedding` layer can also be invoked directly.  Explicit
    preprocessing is required when used with JIT compilation.
    ```python
    # Call the layer directly.
    activations = embedding_layer(my_inputs, my_weights)

    # Call the layer with JIT compilation and explicitly preprocessed inputs.
    embedding_layer_jit = jax.jit(embedding_layer)
    preprocessed_inputs = embedding_layer.preprocess(my_inputs, my_weights)
    activations = embedding_layer_jit(preprocessed_inputs)
    ```

    Similarly, for custom training loops, preprocessing must be applied prior
    to passing the data to the JIT-compiled training step.
    ```python
    # Create an optimizer and loss function.
    optimizer = keras.optimizers.Adam(learning_rate=1e-3)

    def loss_and_updates(trainable_variables, non_trainable_variables, x, y):
        y_pred, non_trainable_variables = model.stateless_call(
            trainable_variables, non_trainable_variables, x, training=True
        )
        loss = keras.losses.mean_squared_error(y, y_pred)
        return loss, non_trainable_variables

    grad_fn = jax.value_and_grad(loss_and_updates, has_aux=True)

    # Create a JIT-compiled training step.
    @jax.jit
    def train_step(state, x, y):
        (
          trainable_variables,
          non_trainable_variables,
          optimizer_variables,
        ) = state
        (loss, non_trainable_variables), grads = grad_fn(
            trainable_variables, non_trainable_variables, x, y
        )
        trainable_variables, optimizer_variables = optimizer.stateless_apply(
            optimizer_variables, grads, trainable_variables
        )
        return loss, (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
        )

    # Build optimizer variables.
    optimizer.build(model.trainable_variables)

    # Assemble the training state.
    trainable_variables = model.trainable_variables
    non_trainable_variables = model.non_trainable_variables
    optimizer_variables = optimizer.variables
    state = trainable_variables, non_trainable_variables, optimizer_variables

    # Training loop.
    for (inputs, weights), labels in train_dataset:
        # Explicitly preprocess the data.
        preprocessed_inputs = embedding_layer.preprocess(inputs, weights)
        loss, state = train_step(state, preprocessed_inputs, labels)
    ```

    Args:
        feature_configs: A nested structure of `keras_rs.layers.FeatureConfig`.
        table_stacking: The table stacking to use. `None` means no table
            stacking. `"auto"` means to stack tables automatically. A list of
            table names or list of lists of table names means to stack the
            tables in the inner lists together. Note that table stacking is not
            supported on older TPUs, in which case the default value of `"auto"`
            will be interpreted as no table stacking.
        **kwargs: Additional arguments to pass to the layer base class.
    """

    def __init__(
        self,
        feature_configs: types.Nested[FeatureConfig],
        *,
        table_stacking: (
            str | Sequence[str] | Sequence[Sequence[str]]
        ) = "auto",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        self._init_feature_configs_structures(feature_configs)

        # Initialize for features placed on "sparsecore".
        if "sparsecore" in self._placement_to_path_to_feature_config:
            self._sparsecore_init(
                self._placement_to_path_to_feature_config["sparsecore"],
                table_stacking,
            )
        # Initialize for features placed on "default_device".
        if "default_device" in self._placement_to_path_to_feature_config:
            self._default_device_init(
                self._placement_to_path_to_feature_config["default_device"],
                table_stacking,
            )

    @keras_utils.no_automatic_dependency_tracking
    def _init_feature_configs_structures(
        self,
        feature_configs: types.Nested[FeatureConfig],
    ) -> None:
        """Initializations for efficiently transforming nested structures.

        This layer handles arbitrarily nested structures for input features, and
        therefore for outputs and feature configs. However, as an intermediary
        format we use a two-level representation with nested dicts. the top
        level dict is keyed by placement and the inner dict is keyed by path,
        with the path representing the path in the original deeply nested
        structure. Thanks to this intermediate representation, we can:
        - dispatch the inputs by placement to overridden methods
        - have backend specific implementations support only one level of
          nesting.

        This method is responsible for creating structures that allow this
        conversion to happen in a few lines of code and efficiently. The
        following attributes are created:
        - self._feature_configs: the deeply nested `FeatureConfig` instances as
          provided by user in `__init__`
        - self._feature_deeply_nested_placement_and_paths: `PlacementAndPath`
          instances in the same deeply nested structure as
          `self._feature_configs`. Needed for `build` because flatten cannot be
          used as it would expand the shape tuples.
        - self._placement_to_path_to_feature_config: `FeatureConfig` instances
          in the same two-level representation keyed by placement and then path.
          Used to go from a flat representation to the intermediate
          representation.

        With these structures in place, the steps to:
        - go from the deeply nested structure to the two-level structure are:
          - `assert_same_struct` as `self._feature_configs`
          - use `self._feature_deeply_nested_placement_and_paths` to map from
            deeply nested to two-level
        - go from the two-level structure to the deeply nested structure:
          - `assert_same_struct` as `self._placement_to_path_to_feature_config`
          - use `self._feature_deeply_nested_placement_and_paths` to locate each
            output in the two-level dicts

        Args:
            feature_configs: The deeply nested structure of `FeatureConfig` or
                `tf.tpu.experimental.embedding.FeatureConfig` as provided by the
                user.
        """
        # Needs to be assigned with `no_automatic_dependency_tracking` to not
        # alter the data structure types.
        self._feature_configs = feature_configs

        placement_and_paths: list[PlacementAndPath] = []
        paths_and_feature_configs = keras.tree.flatten_with_path(
            self._feature_configs
        )
        self._placement_to_path_to_feature_config: dict[
            str, dict[str, FeatureConfig]
        ] = {}

        # Lazily initialized.
        has_sparsecore = None

        for path, feature_config in paths_and_feature_configs:
            if isinstance(feature_config, FeatureConfig):
                placement = feature_config.table.placement
                # Resolve "auto" to an actual placement.
                if placement == "auto":
                    if has_sparsecore is None:
                        has_sparsecore = self._has_sparsecore()
                    placement = (
                        "sparsecore" if has_sparsecore else "default_device"
                    )
            else:
                # It's a `tf.tpu.experimental.embedding.FeatureConfig`.
                placement = "sparsecore"

            path = ".".join([str(e) for e in path])
            if placement not in SUPPORTED_PLACEMENTS:
                raise ValueError(
                    f"Feature '{path}' with name '{feature_config.name}' has "
                    f"unsupported placement '{placement}'."
                )
            placement_and_paths.append(PlacementAndPath(placement, path))
            if placement not in self._placement_to_path_to_feature_config:
                self._placement_to_path_to_feature_config[placement] = {}
            self._placement_to_path_to_feature_config[placement][path] = (
                feature_config
            )

        self._feature_deeply_nested_placement_and_paths = (
            keras.tree.pack_sequence_as(
                self._feature_configs, placement_and_paths
            )
        )

    def build(self, input_shapes: types.Nested[types.Shape]) -> None:
        if self.built:
            return

        self._verify_input_shapes(input_shapes)

        # Go from deeply nested structure to placement -> path -> input shape.
        placement_to_path_to_input_shape: collections.defaultdict[
            str, dict[str, types.Shape]
        ] = collections.defaultdict(dict)

        def populate_placement_to_path_to_input_shape(
            pp: PlacementAndPath, input_shape: types.Shape
        ) -> None:
            placement_to_path_to_input_shape[pp.placement][pp.path] = (
                input_shape
            )

        keras.tree.map_structure_up_to(
            self._feature_deeply_nested_placement_and_paths,
            populate_placement_to_path_to_input_shape,
            self._feature_deeply_nested_placement_and_paths,
            input_shapes,
        )

        # Build for features placed on "sparsecore".
        if "sparsecore" in placement_to_path_to_input_shape:
            self._sparsecore_build(
                placement_to_path_to_input_shape["sparsecore"]
            )

        # Build for features placed on "default_device".
        if "default_device" in placement_to_path_to_input_shape:
            self._default_device_build(
                placement_to_path_to_input_shape["default_device"]
            )

        super().build(input_shapes)

    def preprocess(
        self,
        inputs: types.Nested[types.Tensor],
        weights: types.Nested[types.Tensor] | None = None,
        training: bool = False,
    ) -> types.Nested[types.Tensor]:
        """Preprocesses and reformats the data for consumption by the model.

        For the JAX backend, converts the input data to a hardward-dependent
        format required for use with SparseCores.  Calling `preprocess`
        explicitly is only necessary to enable `jit_compile = True`.

        For non-JAX backends, preprocessing will bundle together the inputs and
        weights, and separate the inputs by device placement.  This step is
        entirely optional.

        Args:
            inputs: Ragged or dense set of sample IDs.
            weights: Optional ragged or dense set of sample weights.
            training: If true, will update internal parameters, such as
                required buffer sizes for the preprocessed data.

        Returns:
            Set of preprocessed inputs that can be fed directly into the
            `inputs` argument of the layer.
        """
        # Verify input structure.
        keras.tree.assert_same_structure(self._feature_configs, inputs)
        if weights is not None:
            keras.tree.assert_same_structure(self._feature_configs, weights)

        if not self.built:
            input_shapes = keras.tree.map_structure(
                lambda array: backend.standardize_shape(array.shape),
                inputs,
            )
            self.build(input_shapes)

        # Go from deeply nested to nested dict placement -> path -> input.
        def to_placement_to_path(
            tensors: types.Nested[types.Tensor],
        ) -> dict[str, dict[str, types.Tensor]]:
            result: dict[str, dict[str, types.Tensor]] = {
                p: dict() for p in self._placement_to_path_to_feature_config
            }

            def populate(pp: PlacementAndPath, x: types.Tensor) -> None:
                result[pp.placement][pp.path] = x

            keras.tree.map_structure(
                populate,
                self._feature_deeply_nested_placement_and_paths,
                tensors,
            )
            return result

        placement_to_path_to_inputs = to_placement_to_path(inputs)

        # Same for weights if present.
        placement_to_path_to_weights = (
            to_placement_to_path(weights) if weights is not None else None
        )

        placement_to_path_to_preprocessed: dict[
            str, dict[str, dict[str, types.Nested[types.Tensor]]]
        ] = {}

        # Preprocess for features placed on "sparsecore".
        if "sparsecore" in placement_to_path_to_inputs:
            placement_to_path_to_preprocessed["sparsecore"] = (
                self._sparsecore_preprocess(
                    placement_to_path_to_inputs["sparsecore"],
                    placement_to_path_to_weights["sparsecore"]
                    if placement_to_path_to_weights is not None
                    else None,
                    training,
                )
            )

        # Preprocess for features placed on "default_device".
        if "default_device" in placement_to_path_to_inputs:
            placement_to_path_to_preprocessed["default_device"] = (
                self._default_device_preprocess(
                    placement_to_path_to_inputs["default_device"],
                    placement_to_path_to_weights["default_device"]
                    if placement_to_path_to_weights is not None
                    else None,
                    training,
                )
            )

        # Mark inputs as preprocessed using an extra level of nesting.
        # This is necessary to detect whether inputs are already preprocessed
        # in `call`.
        output = {
            "preprocessed_inputs_per_placement": (
                placement_to_path_to_preprocessed
            )
        }
        return output

    def _is_preprocessed(
        self, inputs: types.Nested[types.Tensor | types.Shape]
    ) -> bool:
        """Checks if the input is already preprocessed."""
        return (
            isinstance(inputs, dict)
            and "preprocessed_inputs_per_placement" in inputs
        )

    def call(
        self,
        inputs: types.Nested[types.Tensor],
        weights: types.Nested[types.Tensor] | None = None,
        training: bool = False,
    ) -> types.Nested[types.Tensor]:
        """Lookup features in embedding tables and apply reduction.

        Args:
            inputs: A nested structure of 2D tensors to embed and reduce. The
                structure must be the same as the `feature_configs` passed
                during construction.  Alternatively, may consist of already
                preprocessed inputs (see `preprocess`).
            weights: An optional nested structure of 2D tensors of weights to
               apply before reduction. When present, the structure must be the
               same as `inputs` and the shapes must match.
            training: Whether we are training or evaluating the model.

        Returns:
            A nested structure of dense 2D tensors, which are the reduced
            embeddings from the passed features. The structure is the same as
            `inputs`.
        """
        preprocessed_inputs = inputs
        # Preprocess if not already done.
        if not self._is_preprocessed(inputs):
            preprocessed_inputs = self.preprocess(inputs, weights, training)

        preprocessed_inputs = typing.cast(
            dict[str, dict[str, dict[str, types.Tensor]]], preprocessed_inputs
        )
        # Placement -> path -> preprocessed inputs.
        preprocessed_inputs = preprocessed_inputs[
            "preprocessed_inputs_per_placement"
        ]

        placement_to_path_to_outputs = {}

        # Call for features placed on "sparsecore".
        if "sparsecore" in preprocessed_inputs:
            inputs_and_weights = preprocessed_inputs["sparsecore"]
            placement_to_path_to_outputs["sparsecore"] = self._sparsecore_call(
                **inputs_and_weights,
                training=training,
            )

        # Call for features placed on "default_device".
        if "default_device" in preprocessed_inputs:
            inputs_and_weights = preprocessed_inputs["default_device"]
            placement_to_path_to_outputs["default_device"] = (
                self._default_device_call(
                    **inputs_and_weights,
                    training=training,
                )
            )

        # Verify output structure.
        keras.tree.assert_same_structure(
            self._placement_to_path_to_feature_config,
            placement_to_path_to_outputs,
        )

        # Go from placement -> path -> output to deeply nested structure.
        def populate_output(pp: PlacementAndPath) -> types.Tensor:
            return placement_to_path_to_outputs[pp.placement][pp.path]

        return keras.tree.map_structure(
            populate_output, self._feature_deeply_nested_placement_and_paths
        )

    def get_embedding_tables(self) -> dict[str, types.Tensor]:
        """Return the content of the embedding tables by table name.

        The tables are keyed by the name provided in each `TableConfig`. Note
        that the returned tensors are not the actual embedding table variables
        used internally by `DistributedEmbedding`.

        Returns:
            A dictionary of table name to tensor for the embedding tables.
        """
        tables = {}
        if "sparsecore" in self._placement_to_path_to_feature_config:
            tables.update(self._sparsecore_get_embedding_tables())
        if "default_device" in self._placement_to_path_to_feature_config:
            tables.update(self._default_device_get_embedding_tables())
        return tables

    def _default_device_init(
        self,
        feature_configs: dict[str, FeatureConfig],
        table_stacking: str | Sequence[Sequence[str]],
    ) -> None:
        del table_stacking
        table_config_id_to_embedding_layer: dict[int, EmbedReduce] = {}
        self._default_device_embedding_layers: dict[str, EmbedReduce] = {}

        for path, feature_config in feature_configs.items():
            if id(feature_config.table) in table_config_id_to_embedding_layer:
                self._default_device_embedding_layers[path] = (
                    table_config_id_to_embedding_layer[id(feature_config.table)]
                )
            else:
                embedding_layer = EmbedReduce(
                    name=feature_config.table.name,
                    input_dim=feature_config.table.vocabulary_size,
                    output_dim=feature_config.table.embedding_dim,
                    embeddings_initializer=feature_config.table.initializer,
                    combiner=feature_config.table.combiner,
                )
                table_config_id_to_embedding_layer[id(feature_config.table)] = (
                    embedding_layer
                )
                self._default_device_embedding_layers[path] = embedding_layer

    def _default_device_build(
        self, input_shapes: dict[str, types.Shape]
    ) -> None:
        for path, input_shape in input_shapes.items():
            embedding_layer = self._default_device_embedding_layers[path]
            if not embedding_layer.built:
                embedding_layer.build(input_shape)

    def _default_device_preprocess(
        self,
        inputs: dict[str, types.Tensor],
        weights: dict[str, types.Tensor] | None,
        training: bool = False,
    ) -> dict[str, dict[str, types.Tensor]]:
        del training

        # NOTE: This JAX specialization is in the base layer so it is available
        #       on all platforms.  The superclass jax.DistributedEmbedding layer
        #       is currently only imported in linux_x86_64.
        if keras.backend.backend() == "jax":
            feature_configs = self._placement_to_path_to_feature_config[
                "default_device"
            ]

            # Potentially track new weights.  For ragged inputs, if we
            # densify, we will generate a dense weight tensor.
            new_weights: dict[str, types.Tensor] = {}
            use_weights = weights is not None

            # Convert any ragged inputs to dense.
            for path, config in feature_configs.items():
                feature_inputs = inputs[path]
                feature_weights = weights[path] if weights is not None else None

                feature_valence = (
                    None
                    if len(config.input_shape) <= 1
                    else config.input_shape[1]
                )
                feature_inputs, feature_weights = _ragged_to_dense_inputs(
                    feature_inputs, feature_weights, feature_valence
                )
                # Converting to ragged may have introduced a weights array.
                use_weights = use_weights or feature_weights is not None
                inputs[path] = feature_inputs
                new_weights[path] = feature_weights

            if use_weights:
                weights = new_weights

        output: dict[str, types.Tensor] = {"inputs": inputs}
        if weights is not None:
            output["weights"] = weights

        return output

    def _default_device_call(
        self,
        inputs: dict[str, types.Tensor],
        weights: dict[str, types.Tensor] | None = None,
        training: bool = False,
    ) -> dict[str, types.Tensor]:
        del training  # Unused by default.
        if weights is None:
            return {
                path: self._default_device_embedding_layers[path](x)
                for path, x in inputs.items()
            }
        else:
            return {
                path: self._default_device_embedding_layers[path](
                    x, weights[path]
                )
                for path, x in inputs.items()
            }

    def _default_device_get_embedding_tables(self) -> dict[str, types.Tensor]:
        tables = {}
        for path, feature_config in self._placement_to_path_to_feature_config[
            "default_device"
        ].items():
            tables[feature_config.table.name] = (
                self._default_device_embedding_layers[path].embeddings.value
            )
        return tables

    def _has_sparsecore(self) -> bool:
        # Explicitly check for SparseCore availability.
        # We need this check here rather than in jax/distributed_embedding.py
        # so that we can warn the user about missing dependencies.
        if keras.backend.backend() == "jax":
            # Check if SparseCores are available.
            try:
                import jax

                tpu_devices = jax.devices("tpu")
            except RuntimeError:
                # No TPUs available.
                return False

            if len(tpu_devices) > 0:
                device_kind = tpu_devices[0].device_kind
                if device_kind in ["TPU v5", "TPU v6 lite"]:
                    return True

        return False

    def _sparsecore_init(
        self,
        feature_configs: dict[str, FeatureConfig],
        table_stacking: str | Sequence[Sequence[str]],
    ) -> None:
        del feature_configs, table_stacking

        if keras.backend.backend() == "jax":
            jax_tpu_embedding_spec = importlib.util.find_spec(
                "jax_tpu_embedding"
            )
            if jax_tpu_embedding_spec is None:
                raise ImportError(
                    "Please install jax-tpu-embedding to use "
                    "DistributedEmbedding on sparsecore devices."
                )

        raise self._unsupported_placement_error("sparsecore")

    def _sparsecore_build(self, input_shapes: dict[str, types.Shape]) -> None:
        del input_shapes
        raise self._unsupported_placement_error("sparsecore")

    def _sparsecore_preprocess(
        self,
        inputs: dict[str, types.Tensor],
        weights: dict[str, types.Tensor] | None,
        training: bool = False,
    ) -> dict[str, dict[str, types.Tensor]]:
        del training
        output: dict[str, types.Tensor] = {"inputs": inputs}
        if weights is not None:
            output["weights"] = weights

        return output

    def _sparsecore_call(
        self,
        inputs: dict[str, types.Tensor],
        weights: dict[str, types.Tensor] | None = None,
        training: bool = False,
    ) -> dict[str, types.Tensor]:
        del inputs, weights, training
        raise self._unsupported_placement_error("sparsecore")

    def _sparsecore_get_embedding_tables(self) -> dict[str, types.Tensor]:
        raise self._unsupported_placement_error("sparsecore")

    def compute_output_shape(
        self, input_shapes: types.Nested[types.Shape]
    ) -> types.Nested[types.Shape]:
        self._verify_input_shapes(input_shapes)
        output_shape: types.Nested[types.Shape] = keras.tree.map_structure(
            lambda fc: fc.output_shape, self._feature_configs
        )
        return output_shape

    def get_config(self) -> dict[str, Any]:
        # Because the Keras serialization creates a tree of serialized objects,
        # it does not directly support sharing tables between feature configs.
        # We therefore serialize the tables config as a flat list and then refer
        # to them by index in each feature config.

        # The serialized `TableConfig` objects.
        table_config_dicts: list[dict[str, Any]] = []
        # Mapping from `TableConfig` id to index in `table_config_dicts`.
        table_config_id_to_index: dict[int, int] = {}

        def serialize_feature_config(
            feature_config: FeatureConfig,
        ) -> dict[str, Any]:
            # Note that for consistency with the contract of `get_config`, the
            # returned dict contains the serialized `TableConfig` in the "table"
            # key.
            feature_config_dict = feature_config.get_config()

            if id(feature_config.table) not in table_config_id_to_index:
                # Save the serialized `TableConfig` the first time we see it and
                # remember its index.
                table_config_id_to_index[id(feature_config.table)] = len(
                    table_config_dicts
                )
                table_config_dicts.append(feature_config_dict["table"])

            # Replace the serialized `TableConfig` with its index.
            feature_config_dict["table"] = table_config_id_to_index[
                id(feature_config.table)
            ]
            return feature_config_dict

        config: dict[str, Any] = super().get_config()
        config["feature_configs"] = keras.tree.map_structure(
            serialize_feature_config, self._feature_configs
        )
        config["tables"] = table_config_dicts
        if hasattr(self, "_table_stacking"):
            config["table_stacking"] = self._table_stacking
        return config

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "DistributedEmbedding":
        config = config.copy()
        # We need to reconnect the `TableConfig`s to the `FeatureConfig`s.

        # The serialized `TableConfig` objects.
        table_config_dicts: list[dict[str, Any]] = config.pop("tables")
        # The deserialized `TableConfig` objects at the same indices.
        table_configs: list[TableConfig | None] = [None] * len(
            table_config_dicts
        )

        def deserialize_feature_config(
            feature_config_dict: dict[str, Any],
        ) -> FeatureConfig | None:
            # Look for a "name" attribute which is a string to detect a
            # `FeatureConfig` leaf node. If not, keep recursing.
            if "name" not in feature_config_dict or not isinstance(
                feature_config_dict["name"], str
            ):
                # Tell `traverse` to recurse.
                return None

            table_index = feature_config_dict["table"]
            # Note that for consistency with the contract of `from_config`, the
            # passed dict must contain the serialized `TableConfig` in the
            # "table" key.
            feature_config_dict["table"] = table_config_dicts[table_index]
            feature_config = FeatureConfig.from_config(feature_config_dict)
            # But then dedupe `TableConfig`s.
            if table_configs[table_index] is None:
                # Remember each new `TableConfig` we see.
                table_configs[table_index] = feature_config.table
            else:
                # And swap duplicates for the original.
                feature_config.table = table_configs[table_index]
            return feature_config

        # Because each `FeatureConfig` is serialized as a dict, we cannot use
        # `map_structure` as it would recurse in the config itself. We use
        # `traverse` instead with a function that detects leaf nodes.
        config["feature_configs"] = keras.tree.traverse(
            deserialize_feature_config, config["feature_configs"]
        )
        return cls(**config)

    def _verify_input_shapes(
        self, input_shapes: types.Nested[types.Shape]
    ) -> None:
        """Verifies that the input shapes match the ones in the feature configs.

        Args:
          input_shapes: The structure of input shapes to verify.
        """
        # Support preprocessing.
        if self._is_preprocessed(input_shapes):
            # Structure should be :
            # {
            #   placement: {
            #     inputs: {path: Any},
            #     weights: {path: Any}
            #   }
            # }
            #
            # But the `Any` values could be nested tensors with varying
            # structure, depending on hardware constraints.  This complicates
            # checking shapes via keras.tree methods.  So, assume the
            # input is a result of explicitly calling the `preprocess(...)`
            # function, in which case the structure has already been verified.
            return

        def _verify_input_shape(
            feature_config: FeatureConfig,
            input_shape: types.Shape,
        ) -> None:
            if not isinstance(input_shape, (tuple, list)) or not all(
                isinstance(d, (int, type(None))) for d in input_shape
            ):
                raise ValueError(f"Received invalid input shape {input_shape}.")
            if len(input_shape) < 1:
                raise ValueError(
                    f"Received input shape {input_shape}. Rank must be 1 or "
                    "above."
                )
            keras_utils.check_shapes_compatible(
                feature_config.input_shape, input_shape
            )

        keras.tree.map_structure_up_to(
            self._feature_configs,
            _verify_input_shape,
            self._feature_configs,
            input_shapes,
        )

    def _unsupported_placement_error(self, placement: str) -> Exception:
        return NotImplementedError(
            f"Backend '{keras.backend.backend()}' does not support the "
            f"'{placement}' placement."
        )



from typing import Any

import keras
from keras import ops

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export
from keras_rs.src.utils.keras_utils import check_shapes_compatible

SUPPORTED_COMBINERS = ("mean", "sum", "sqrtn")


def _is_supported_sparse(x: types.Tensor) -> bool:
    """Determines if the input is a supported sparse tensor.

    NOTE: Currently only works for the TensorFlow and JAX backends.

    Args:
      x: Input tensor to check for sparsity.

    Returns:
      True if `x` is a supported sparse tensor.
    """
    if keras.backend.backend() == "tensorflow":
        import tensorflow as tf

        return isinstance(x, tf.SparseTensor)
    elif keras.backend.backend() == "jax":
        from jax.experimental import sparse as jax_sparse

        return isinstance(x, jax_sparse.BCOO) or isinstance(x, jax_sparse.BCSR)

    return False


def _sparse_ones_like(
    x: types.Tensor, dtype: types.DType | None = None
) -> types.Tensor:
    """Creates a tensor of ones with the same sparsity as the input.

    This differs from `keras.ops.ones_like`, which would create a dense
    tensor of ones.

    Args:
        x: Input sparse tensor.
        dtype: Optional dtype for the output tensor values.

    Returns:
        Sparse tensor of ones.

    Raises:
        ValueError for unsupported sparse input type and backend.
    """
    dtype = dtype or x.dtype
    if keras.backend.backend() == "tensorflow":
        import tensorflow as tf

        # Ensure shape is copied exactly for compatibility in graph mode.
        x_shape = x.shape
        y = tf.SparseTensor(
            x.indices, tf.ones_like(x.values, dtype=dtype), x.dense_shape
        )
        y.set_shape(x_shape)
        return y
    elif keras.backend.backend() == "jax":
        import jax.numpy as jnp
        from jax.experimental import sparse as jax_sparse

        if isinstance(x, jax_sparse.BCOO):
            return jax_sparse.BCOO(
                (jnp.ones_like(x.data, dtype=dtype), x.indices),
                shape=x.shape,
                indices_sorted=x.indices_sorted,
                unique_indices=x.unique_indices,
            )
        elif isinstance(x, jax_sparse.BCSR):
            return jax_sparse.BCSR(
                (jnp.ones_like(x.data, dtype=dtype), x.indices, x.indptr),
                shape=x.shape,
                indices_sorted=x.indices_sorted,
                unique_indices=x.unique_indices,
            )

    raise ValueError(
        f"Unsupported sparse input type '{x.__class__.__name__}' for backend "
        f"{keras.backend.backend()}."
    )


@keras_rs_export("keras_rs.layers.EmbedReduce")
class EmbedReduce(keras.layers.Embedding):
    """An embedding layer that reduces with a combiner.

    This layer embeds inputs and then applies a reduction to combine a set of
    embeddings into a single embedding. This is typically used to embed a
    sequence of items as a single embedding.

    If the inputs passed to `__call__` are 1D, no reduction is applied. If the
    inputs are 2D, dimension 1 is reduced using the combiner so that the result
    is of shape `(batch_size, output_dim`). Inputs of rank 3 and higher are not
    allowed. Weights can optionally be passed to the `__call__` method to
    apply weights to different samples before reduction.

    This layer supports sparse inputs and ragged inputs with backends that
    support them. The output after reduction is dense. For ragged inputs, the
    ragged dimension must be 1 as it is the dimension that is reduced.

    Args:
        input_dim: Integer. Size of the vocabulary, maximum integer index + 1.
        output_dim: Integer. Dimension of the dense embedding.
        embeddings_initializer: Initializer for the `embeddings` matrix (see
            `keras.initializers`).
        embeddings_regularizer: Regularizer function applied to the `embeddings`
            matrix (see `keras.regularizers`).
        embeddings_constraint: Constraint function applied to the `embeddings`
            matrix (see `keras.constraints`).
        mask_zero: Boolean, whether or not the input value 0 is a special
            "padding" value that should be masked out. This is useful when using
            recurrent layers which may take variable length input. If this is
            `True`, then all subsequent layers in the model need to support
            masking or an exception will be raised. If `mask_zero` is set to
            `True`, as a consequence, index 0 cannot be used in the vocabulary
            (`input_dim` should equal size of vocabulary + 1).
        weights: Optional floating-point matrix of size
            `(input_dim, output_dim)`. The initial embeddings values to use.
        combiner: Specifies how to reduce if there are multiple entries in a
            single row. Currently `mean`, `sqrtn` and `sum` are supported.
            `mean` is the default. `sqrtn` often achieves good accuracy, in
            particular with bag-of-words columns.
        **kwargs: Additional keyword arguments passed to `Embedding`.
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        embeddings_initializer: types.InitializerLike = "uniform",
        embeddings_regularizer: types.RegularizerLike | None = None,
        embeddings_constraint: types.ConstraintLike | None = None,
        mask_zero: bool = False,
        weights: types.Tensor = None,
        combiner: str = "mean",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            input_dim,
            output_dim,
            embeddings_initializer=embeddings_initializer,
            embeddings_regularizer=embeddings_regularizer,
            embeddings_constraint=embeddings_constraint,
            mask_zero=mask_zero,
            weights=weights,
            **kwargs,
        )
        if combiner not in SUPPORTED_COMBINERS:
            raise ValueError(
                f"Invalid `combiner`: '{combiner}', "
                f"use one of {', '.join(SUPPORTED_COMBINERS)}."
            )
        self.combiner = combiner

    def call(
        self,
        inputs: types.Tensor,
        weights: types.Tensor | None = None,
    ) -> types.Tensor:
        """Apply embedding and reduction.

        Args:
            inputs: 1D tensor to embed or 2D tensor to embed and reduce.
            weights: Optional tensor of weights to apply before reduction, which
               can be 1D or 2D and must match for the first dimension of
               `inputs` (1D case) or match the shape of `inputs` (2D case).

        Returns:
            A dense 2D tensor of shape `(batch_size, output_dim)`.
        """
        x = super().call(inputs)
        unreduced_rank = len(x.shape)

        # Check that weights has a compatible shape.
        if weights is not None:
            weights_rank = len(weights.shape)
            if weights_rank > unreduced_rank or not check_shapes_compatible(
                x.shape[0:weights_rank], weights.shape
            ):
                raise ValueError(
                    f"The shape of `weights`: {weights.shape} is not compatible"
                    f" with the shape of `inputs` after embedding: {x.shape}."
                )

        dtype = (
            x.dtype
            if weights is None
            else keras.backend.result_type(x.dtype, weights.dtype)
        )

        # When `weights` is `None`:
        # - For ragged inputs, after embedding, we get a ragged result that has
        #   a ragged dimension of 1, but when we do the "mean" or "sqrtn", we
        #   need to divide by the number of items in each row. However, there is
        #   no explicit cross backend API to get the row length. `ones_like`
        #   gives us a ragged tensor that is ragged in the same way as the
        #   inputs. When we do `ops.sum(weights, axis=-2)`, it gives us the
        #   number of items per row.
        # - For sparse inputs, after embedding, we get a dense tensor, not a
        #   sparse tensor. What it does for missing values is use embedding 0.
        #   These are bogus embedding and should be ignored. `ones_like` gives
        #   us a sparse tensor with the exact same missing values. Later, when
        #   we do `x = ops.multiply(x, weights)`, which masks the bogus values
        #   (note that `weights` has been densified beforehand). Additionally,
        #   when we do `ops.sum(weights, axis=-2)`, it gives us the number of
        #   items per row.
        #
        # When `unreduced_rank <= 2`, this means that the inputs where 1D and
        # dense, there is only one embedding per row, so there is no real
        # reduction is going on.
        # - For mean: result = weights * x / weights = x we don't need `weights`
        # - For sqrtn: result = weights * x / sqrt(square(weights)) = x we don't
        #   needs `weights`
        # - For sum however: `result = weights * x` we do need `weights`.
        # So for mean and sqrtn we don't need the weights, we use ones instead.
        # This is to avoid divisions by zero and improve the precision.
        if weights is None or (unreduced_rank <= 2 and self.combiner != "sum"):
            # Discard the weights if there were some and create a mask for
            # ragged and sparse tensors to mask the result correctly (sparse
            # only) and the apply the reduction correctly (ragged and sparse).
            if _is_supported_sparse(inputs):
                weights = _sparse_ones_like(inputs, dtype=dtype)
            else:
                weights = ops.ones_like(inputs, dtype=dtype)

        else:
            weights = ops.cast(weights, dtype)

        # When looking up using sparse indices, the result is dense but contains
        # values that should be ignored as all missing values use index 0. We
        # use `weights` as a mask, but it needs to be densified as
        # `expand_dims` and broadcasting a sparse tensor does not produce the
        # expected result.
        weights = ops.convert_to_tensor(weights, sparse=False)

        # Make weights and the unreduced embeddings have the same rank.
        weights_rank = len(weights.shape)
        if weights_rank < unreduced_rank:
            weights = ops.expand_dims(
                weights, axis=tuple(range(weights_rank, unreduced_rank))
            )

        # Note that `x` and `weights` are:
        # - ragged if `inputs` was ragged and `weights` was ragged or None
        # - dense otherwise (even if `inputs` and `weights` were sparse).
        x = ops.multiply(x, weights)

        if unreduced_rank <= 2:
            # No reduction is applied.
            return x

        # After this reduction, `x` is always dense as we reduce the ragged
        # dimension in the ragged case.
        x = ops.sum(x, axis=-2)

        # Apply the right divisor for the combiner.
        # Where we use `weights` in the divisor, we use
        # `ops.sum(weights, axis=-2)` which always makes it dense as we reduce
        # the ragged dimension in the ragged case.
        if self.combiner == "mean":
            return ops.divide_no_nan(x, ops.sum(weights, axis=-2))
        elif self.combiner == "sum":
            return x
        elif self.combiner == "sqrtn":
            return ops.divide_no_nan(
                x, ops.sqrt(ops.sum(ops.square(weights), axis=-2))
            )

    def compute_output_shape(
        self,
        input_shape: types.Shape,
        weights_shape: types.Shape | None = None,
    ) -> types.Shape:
        del weights_shape

        if len(input_shape) <= 1:
            # No reduction
            return (*input_shape, self.output_dim)
        else:
            # Reduce last dimension
            return (*input_shape[0:-1], self.output_dim)

    def compute_output_spec(
        self,
        inputs: keras.KerasTensor,
        weights: keras.KerasTensor | None = None,
    ) -> keras.KerasTensor:
        del weights

        output_shape = self.compute_output_shape(inputs.shape)
        return keras.KerasTensor(output_shape, dtype=self.compute_dtype)

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = super().get_config()

        config.update(
            {
                "combiner": self.combiner,
            }
        )

        return config



"""Defines a differentiable `embedding_lookup` function.

Implementation details for use in JAX models.
"""

import functools
from typing import Any, Mapping, TypeAlias

import jax
import numpy as np
from jax.experimental import layout as jax_layout
from jax_tpu_embedding.sparsecore.lib.nn import embedding
from jax_tpu_embedding.sparsecore.lib.nn import embedding_spec
from jax_tpu_embedding.sparsecore.utils import utils as jte_utils

from keras_rs.src.layers.embedding.jax import embedding_utils
from keras_rs.src.types import Nested

ShardedCooMatrix = embedding_utils.ShardedCooMatrix
shard_map = jax.experimental.shard_map.shard_map  # type: ignore[attr-defined]

ArrayLike: TypeAlias = jax.Array | np.ndarray[Any, Any]
JaxLayout: TypeAlias = jax.sharding.NamedSharding | jax_layout.Format


class EmbeddingLookupConfiguration:
    """Feature, mesh and sharding configrmation for lookups."""

    mesh: jax.sharding.Mesh
    feature_specs: embedding.Nested[embedding_spec.FeatureSpec]
    table_sharding_strategy: str
    num_sc_per_device: int
    samples_partition: jax.sharding.PartitionSpec
    table_partition: jax.sharding.PartitionSpec
    table_layout: JaxLayout

    def __init__(
        self,
        feature_specs: embedding.Nested[embedding_spec.FeatureSpec],
        mesh: jax.sharding.Mesh | None = None,
        table_sharding_strategy: str = "MOD",
        num_sc_per_device: int | None = None,
        sharding_axis: str = "sparsecore_sharding",
        samples_partition: jax.sharding.PartitionSpec | None = None,
        samples_layout: JaxLayout | None = None,
        table_partition: jax.sharding.PartitionSpec | None = None,
        table_layout: JaxLayout | None = None,
    ):
        self.mesh = mesh or jax.sharding.Mesh(jax.devices(), sharding_axis)
        self.feature_specs = feature_specs
        self.table_sharding_strategy = table_sharding_strategy
        self.num_sc_per_device = (
            num_sc_per_device
            if num_sc_per_device is not None
            else jte_utils.num_sparsecores_per_device()
        )
        self.samples_partition = (
            samples_partition
            or jax.sharding.PartitionSpec(
                sharding_axis  # type: ignore[no-untyped-call]
            )
        )
        self.samples_layout = samples_layout or jax.sharding.NamedSharding(
            self.mesh, self.samples_partition
        )
        self.table_partition = table_partition or jax.sharding.PartitionSpec(
            sharding_axis,
            None,  # type: ignore[no-untyped-call]
        )
        self.table_layout = table_layout or jax.sharding.NamedSharding(
            self.mesh, self.table_partition
        )


# Embedding lookup function with custom gradient.
@functools.partial(jax.custom_vjp, nondiff_argnums=(0,))
def embedding_lookup(
    config: EmbeddingLookupConfiguration,
    lookups: Mapping[str, ShardedCooMatrix],
    tables: Nested[jax.Array],
    step: jax.Array | None = None,
) -> Nested[jax.Array]:
    """Embedding lookup function with custom gradient.

    Args:
      config: Embedding lookup configuration.
      lookups: Embedding lookup stacked/sharded inputs.
      tables: Embedding lookup stacked/sharded tables.
      step: Current training step number.
    """
    del step  # Only used in backward pass.

    # Decompose COO matrices.
    row_pointers_raw = {}
    embedding_ids_raw = {}
    sample_ids_raw = {}
    gains_raw = {}
    for table_name, coo in lookups.items():
        row_pointers_raw[table_name] = coo.shard_ends
        embedding_ids_raw[table_name] = coo.col_ids
        sample_ids_raw[table_name] = coo.row_ids
        gains_raw[table_name] = coo.values

    sparse_dense_matmul_input = embedding.SparseDenseMatmulInput(
        row_pointers_raw,
        embedding_ids_raw,
        sample_ids_raw,
        gains_raw,
    )

    pd = config.samples_partition
    pt = config.table_partition
    sharded_matmul = jax.jit(
        shard_map(
            functools.partial(
                embedding.tpu_sparse_dense_matmul,
                global_device_count=config.mesh.shape[pd[0]],
                feature_specs=config.feature_specs,
                sharding_strategy=config.table_sharding_strategy,
            ),
            mesh=config.mesh,
            in_specs=(pd, pt),
            out_specs=pd,
            check_rep=False,
        ),
    )

    activations: Nested[jax.Array] = sharded_matmul(
        sparse_dense_matmul_input,
        tables,
    )

    return activations


def embedding_lookup_fwd(
    config: EmbeddingLookupConfiguration,
    lookups: Mapping[str, ShardedCooMatrix],
    table: Nested[jax.Array],
    step: jax.Array | None = None,
) -> tuple[
    Nested[jax.Array],
    tuple[Nested[ShardedCooMatrix], Nested[jax.Array], jax.Array | None],
]:
    """Forward pass for embedding lookup."""
    return embedding_lookup(config, lookups, table, step), (
        lookups,
        table,
        step,
    )


def embedding_lookup_bwd(
    config: EmbeddingLookupConfiguration,
    res: tuple[
        Mapping[str, ShardedCooMatrix],  # Lookups.
        Mapping[str, Nested[jax.Array]],  # Tables.
        jax.Array | None,  # Step.
    ],
    gradients: Nested[jax.Array],
) -> tuple[None, Nested[jax.Array], jax.Array | None]:
    """Backward pass for embedding lookup.

    Args:
      config: Embedding lookup configuration.
      res: Tuple of embedding lookup (inputs, tables, step).
      gradients: Embedding lookup gradients.

    Returns:
      A tuple of gradients (None, table_grads, step + 1).
    """
    lookups, tables, step = res

    # Decompose COO matrices.
    row_pointers_raw = {}
    embedding_ids_raw = {}
    sample_ids_raw = {}
    gains_raw = {}
    for table_name, coo in lookups.items():
        row_pointers_raw[table_name] = coo.shard_ends
        embedding_ids_raw[table_name] = coo.col_ids
        sample_ids_raw[table_name] = coo.row_ids
        gains_raw[table_name] = coo.values

    sparse_dense_matmul_input = embedding.SparseDenseMatmulInput(
        row_pointers_raw,
        embedding_ids_raw,
        sample_ids_raw,
        gains_raw,
    )

    pt = config.table_partition
    pd = config.samples_partition
    # Replicate step count.
    preplicate = jax.sharding.PartitionSpec()  # type: ignore[no-untyped-call]

    def grad_func(
        gradients: Nested[jax.Array],
        sparse_input: embedding.SparseDenseMatmulInput,
        tables: Mapping[str, embedding.EmbeddingVariables],
        step: jax.Array | None,
    ) -> Mapping[str, embedding.EmbeddingVariables]:
        output: Mapping[str, embedding.EmbeddingVariables] = (
            embedding.tpu_sparse_dense_matmul_grad(
                gradients,
                sparse_input,
                tables,
                feature_specs=config.feature_specs,
                sharding_strategy=config.table_sharding_strategy,
                step=step,
            )
        )
        return output

    # activation_layout = jax.sharding.NamedSharding(config.mesh, pd)
    # step_layout = jax.sharding.NamedSharding(config.mesh, preplicate)
    sharded_matmul_grad = jax.jit(
        shard_map(
            grad_func,
            mesh=config.mesh,
            in_specs=(pd, pd, pt, preplicate),
            out_specs=pt,
            check_rep=False,
        ),
        #   in_shardings=(
        #       activation_layout,
        #       config.samples_layout,
        #       config.table_layout,
        #       step_layout,
        #   ),
        # out_shardings=(config.table_layout),
    )

    table_grads = sharded_matmul_grad(
        gradients,
        sparse_dense_matmul_input,
        tables,
        step,
    )

    # tpu_sparse_dense_matmul_grad returns a general Mapping (usually a dict).
    # It may not be the same type as the embedding table (e.g. FrozenDict).
    # Here we use flatten / unflatten to ensure the types are the same.
    table_grads = jax.tree.unflatten(
        jax.tree.structure(tables), jax.tree.leaves(table_grads)
    )

    return (
        None,
        table_grads,
        step + 1 if step is not None else None,  # Incremented step count.
    )


embedding_lookup.defvjp(embedding_lookup_fwd, embedding_lookup_bwd)



"""Utility functions for manipulating JAX embedding tables and inputs."""

import collections
import dataclasses
import typing
from typing import Any, Mapping, NamedTuple, Sequence, TypeAlias, TypeVar

import jax
import numpy as np
from jax import numpy as jnp
from jax_tpu_embedding.sparsecore.lib.nn import embedding
from jax_tpu_embedding.sparsecore.lib.nn.embedding_spec import FeatureSpec
from jax_tpu_embedding.sparsecore.lib.nn.embedding_spec import StackedTableSpec
from jax_tpu_embedding.sparsecore.lib.nn.embedding_spec import TableSpec

from keras_rs.src.types import Nested

T = TypeVar("T")

# Any to support tf.Ragged without needing an explicit TF dependency.
ArrayLike: TypeAlias = jax.Array | np.ndarray | Any  # type: ignore
Shape: TypeAlias = tuple[int, ...]


class FeatureSamples(NamedTuple):
    tokens: ArrayLike
    weights: ArrayLike


class ShardedCooMatrix(NamedTuple):
    shard_starts: ArrayLike
    shard_ends: ArrayLike
    col_ids: ArrayLike
    row_ids: ArrayLike
    values: ArrayLike


class InputStatsPerTable(NamedTuple):
    max_ids_per_partition: int
    max_unique_ids_per_partition: int
    required_buffer_size_per_device: int


def _round_up_to_multiple(value: int, multiple: int) -> int:
    return ((value + multiple - 1) // multiple) * multiple


def _default_stacked_table_spec(
    table_spec: TableSpec, num_shards: int, batch_size: int
) -> StackedTableSpec:
    return StackedTableSpec(
        stack_name=table_spec.name,
        stack_vocab_size=_round_up_to_multiple(
            table_spec.vocabulary_size, 8 * num_shards
        ),
        stack_embedding_dim=_round_up_to_multiple(table_spec.embedding_dim, 8),
        optimizer=table_spec.optimizer,
        combiner=table_spec.combiner,
        total_sample_count=batch_size,
        max_ids_per_partition=table_spec.max_ids_per_partition,
        max_unique_ids_per_partition=table_spec.max_unique_ids_per_partition,
    )


def _get_stacked_table_spec(
    table_spec: TableSpec, num_shards: int, batch_size: int = 0
) -> StackedTableSpec:
    return table_spec.stacked_table_spec or _default_stacked_table_spec(
        table_spec, num_shards, batch_size
    )


def pad_table(
    table_spec: TableSpec,
    table_values: jax.Array,
    num_shards: int,
    pad_value: jnp.float32 = jnp.nan,
) -> jax.Array:
    """Adds appropriate padding to a table to prepare for stacking.

    Args:
        table_spec: Table specification describing the table to pad.
        table_values: Table values array to pad.
        num_shards: Number of shards in the table (typically
            `global_device_count * num_sc_per_device`).
        pad_value: Value to use for padding.

    Returns:
        Padded table values.
    """
    vocabulary_size = table_spec.vocabulary_size
    embedding_dim = table_spec.embedding_dim
    padded_vocabulary_size = _round_up_to_multiple(
        vocabulary_size, 8 * num_shards
    )
    stack_embedding_dim = _get_stacked_table_spec(
        table_spec, num_shards
    ).stack_embedding_dim
    return jnp.pad(
        table_values,
        (
            (0, padded_vocabulary_size - vocabulary_size),
            (0, stack_embedding_dim - embedding_dim),
        ),
        constant_values=pad_value,
    )


def _stack_and_shard_table(
    stacked_table: jax.Array,
    table_spec: TableSpec,
    table: jax.Array,
    num_shards: int,
    pad_value: jnp.float32,
) -> jax.Array:
    """Stacks and shards a single table for use in sparsecore lookups."""
    padded_values = pad_table(table_spec, table, num_shards, pad_value)
    sharded_padded_vocabulary_size = padded_values.shape[0] // num_shards
    stack_embedding_dim = stacked_table.shape[-1]

    # Mod-shard vocabulary across devices.
    sharded_values = jnp.swapaxes(
        padded_values.reshape(-1, num_shards, stack_embedding_dim),
        0,
        1,
    )

    # Rotate shards.
    setting_in_stack = table_spec.setting_in_stack
    rotated_values = jnp.roll(
        sharded_values, setting_in_stack.shard_rotation, axis=0
    )

    # Insert table into the stack.
    table_row = setting_in_stack.row_offset_in_shard
    stacked_table = stacked_table.at[
        :, table_row : (table_row + sharded_padded_vocabulary_size), :
    ].set(rotated_values)

    return stacked_table


def stack_and_shard_tables(
    table_specs: Nested[TableSpec],
    tables: Nested[ArrayLike],
    num_shards: int,
    pad_value: jnp.float32 = jnp.nan,
) -> dict[str, Nested[jax.Array]]:
    """Stacks and shards tables for use in sparsecore lookups.

    Args:
        table_specs: Nested collection of unstacked table specifications.
        tables: Table values corresponding to the table_specs.
        num_shards: Number of shards in the table (typically
            `global_device_count * num_sc_per_device`).
        pad_value: Value to use for padding.

    Returns:
        A mapping of stacked table names to stacked table values.
    """

    # Gather stacked table information.
    stacked_table_map: dict[
        str,
        tuple[StackedTableSpec, list[TableSpec]],
    ] = {}

    def collect_stacked_tables(table_spec: TableSpec) -> None:
        stacked_table_spec = _get_stacked_table_spec(table_spec, num_shards)
        stacked_table_name = stacked_table_spec.stack_name
        if stacked_table_name not in stacked_table_map:
            stacked_table_map[stacked_table_name] = (stacked_table_spec, [])
        stacked_table_map[stacked_table_name][1].append(table_spec)

    _ = jax.tree.map(collect_stacked_tables, table_specs)

    table_map: dict[str, Nested[jax.Array]] = {}

    def collect_tables(table_spec: TableSpec, table: Nested[jax.Array]) -> None:
        table_map[table_spec.name] = table

    _ = jax.tree.map(collect_tables, table_specs, tables)

    stacked_tables: dict[str, Nested[jax.Array]] = {}
    for (
        stacked_table_spec,
        table_specs,
    ) in stacked_table_map.values():
        stack_vocab_size = stacked_table_spec.stack_vocab_size
        sharded_vocab_size = stack_vocab_size // num_shards
        stack_embedding_dim = stacked_table_spec.stack_embedding_dim

        # Allocate initial buffer.  The stacked table will be divided among
        # shards by splitting the vocabulary dimension:
        #   [ v, e ] -> [s, v/s, e]
        stacked_table_tree = jax.tree.map(
            lambda _: jnp.zeros(
                # pylint: disable-next=cell-var-from-loop, used only in loop body.
                shape=(num_shards, sharded_vocab_size, stack_embedding_dim),
                dtype=jnp.float32,
            ),
            table_map[table_specs[0].name],
        )

        for table_spec in table_specs:
            table_tree = table_map[table_spec.name]
            stacked_table_tree = jax.tree.map(
                lambda stacked_table, table: _stack_and_shard_table(
                    # pylint: disable-next=cell-var-from-loop, used only in loop body.
                    stacked_table,
                    # pylint: disable-next=cell-var-from-loop, used only in loop body.
                    table_spec,
                    table,
                    num_shards,
                    pad_value,
                ),
                stacked_table_tree,
                table_tree,
            )

        stacked_tables[stacked_table_spec.stack_name] = stacked_table_tree

    return stacked_tables


def _unshard_and_unstack_table(
    table_spec: TableSpec,
    stacked_table_tree: Nested[jax.Array],
    num_shards: int,
) -> Nested[jax.Array]:
    """Unshards and unstacks a single table."""
    vocabulary_size = table_spec.vocabulary_size
    embedding_dim = table_spec.embedding_dim

    def _unshard_and_unstack_single_table(
        table_spec: TableSpec, stacked_table: jax.Array
    ) -> jax.Array:
        stack_embedding_dim = stacked_table.shape[-1]

        # Maybe re-shape in case it was flattened.
        stacked_table = stacked_table.reshape(
            num_shards, -1, stack_embedding_dim
        )
        sharded_vocabulary_size = (
            _round_up_to_multiple(vocabulary_size, 8 * num_shards) // num_shards
        )

        # Extract padded values from the stacked table.
        setting_in_stack = table_spec.setting_in_stack
        row = setting_in_stack.row_offset_in_shard
        padded_values = stacked_table[
            :, row : (row + sharded_vocabulary_size), :
        ]

        # Un-rotate shards.
        padded_values = jnp.roll(
            padded_values, -setting_in_stack.shard_rotation, axis=0
        )

        # Un-mod-shard.
        padded_values = jnp.swapaxes(padded_values, 0, 1).reshape(
            -1, stack_embedding_dim
        )

        # Un-pad.
        return padded_values[:vocabulary_size, :embedding_dim]

    output: Nested[jax.Array] = jax.tree.map(
        lambda stacked_table: _unshard_and_unstack_single_table(
            table_spec, stacked_table
        ),
        stacked_table_tree,
    )
    return output


def unshard_and_unstack_tables(
    table_specs: Nested[TableSpec],
    stacked_tables: Mapping[str, Nested[jax.Array]],
    num_shards: int,
) -> Nested[jax.Array]:
    """Unshards and unstacks a collection of tables.

    Args:
        table_specs: Nested collection of unstacked table specifications.
        stacked_tables: Mapping of stacked table names to stacked table values.
        num_shards: Number of shards in the table (typically
            `global_device_count * num_sc_per_device`).

    Returns:
        A mapping of table names to unstacked table values.
    """
    output: Nested[jax.Array] = jax.tree.map(
        lambda table_spec: _unshard_and_unstack_table(
            table_spec,
            stacked_tables[
                _get_stacked_table_spec(table_spec, num_shards=1).stack_name
            ],
            num_shards,
        ),
        table_specs,
    )
    return output


def get_table_specs(feature_specs: Nested[FeatureSpec]) -> dict[str, TableSpec]:
    table_spec_map: dict[str, TableSpec] = {}
    flat_feature_specs, _ = jax.tree.flatten(feature_specs)
    for feature_spec in flat_feature_specs:
        table_spec = feature_spec.table_spec
        table_spec_map[table_spec.name] = table_spec
    return table_spec_map


def get_table_stacks(
    table_specs: Nested[TableSpec],
) -> dict[str, list[TableSpec]]:
    """Extracts lists of tables that are stacked together.

    Args:
        table_specs: Nested collection of table specifications.

    Returns:
        A mapping of stacked table names to lists of table specifications for
        each stack.
    """
    stacked_table_specs: dict[str, list[TableSpec]] = collections.defaultdict(
        list
    )
    flat_table_specs, _ = jax.tree.flatten(table_specs)
    for table_spec in flat_table_specs:
        table_spec = typing.cast(TableSpec, table_spec)
        stacked_table_spec = table_spec.stacked_table_spec
        if stacked_table_spec is not None:
            stacked_table_specs[stacked_table_spec.stack_name].append(
                table_spec
            )
        else:
            stacked_table_specs[table_spec.name].append(table_spec)

    return stacked_table_specs


def get_stacked_table_stats(
    feature_specs: Nested[FeatureSpec],
) -> dict[str, InputStatsPerTable]:
    """Extracts the stacked-table input statistics from the feature specs.

    Args:
        feature_specs: Feature specs from which to extracts the statistics.

    Returns:
        A mapping of stacked table names to input statistics per table.
    """
    stacked_table_specs: dict[str, StackedTableSpec] = {}
    for feature_spec in jax.tree.flatten(feature_specs)[0]:
        feature_spec = typing.cast(FeatureSpec, feature_spec)
        stacked_table_spec = typing.cast(
            StackedTableSpec, feature_spec.table_spec.stacked_table_spec
        )
        stacked_table_specs[stacked_table_spec.stack_name] = stacked_table_spec

    stats: dict[str, InputStatsPerTable] = {}
    for stacked_table_spec in stacked_table_specs.values():
        buffer_size = stacked_table_spec.suggested_coo_buffer_size_per_device
        buffer_size = buffer_size or 0
        stats[stacked_table_spec.stack_name] = InputStatsPerTable(
            max_ids_per_partition=stacked_table_spec.max_ids_per_partition,
            max_unique_ids_per_partition=stacked_table_spec.max_unique_ids_per_partition,
            required_buffer_size_per_device=buffer_size,
        )

    return stats


def update_stacked_table_stats(
    feature_specs: Nested[FeatureSpec],
    stats: Mapping[str, InputStatsPerTable],
) -> None:
    """Updates stacked-table input properties in the supplied feature specs.

    Args:
        feature_specs: Feature specs to update in-place.
        stats: Per-stacked-table input statistics.
    """
    # Collect table specs and stacked table specs.
    table_specs: dict[str, TableSpec] = {}
    for feature_spec in jax.tree.flatten(feature_specs)[0]:
        feature_spec = typing.cast(FeatureSpec, feature_spec)
        table_specs[feature_spec.table_spec.name] = feature_spec.table_spec

    stacked_table_specs: dict[str, StackedTableSpec] = {}
    for table_spec in table_specs.values():
        stacked_table_spec = typing.cast(
            StackedTableSpec, table_spec.stacked_table_spec
        )
        stacked_table_specs[stacked_table_spec.stack_name] = stacked_table_spec

    # Replace fields in the stacked_table_specs.
    stack_names = stacked_table_specs.keys()
    for stack_name in stack_names:
        stack_stats = stats[stack_name]
        stacked_table_spec = stacked_table_specs[stack_name]
        buffer_size = stack_stats.required_buffer_size_per_device or None
        stacked_table_specs[stack_name] = dataclasses.replace(
            stacked_table_spec,
            max_ids_per_partition=stack_stats.max_ids_per_partition,
            max_unique_ids_per_partition=stack_stats.max_unique_ids_per_partition,
            suggested_coo_buffer_size_per_device=buffer_size,
        )

    # Insert new stacked tables into tables.
    for table_spec in table_specs.values():
        stacked_table_spec = typing.cast(
            StackedTableSpec, table_spec.stacked_table_spec
        )
        table_spec.stacked_table_spec = stacked_table_specs[
            stacked_table_spec.stack_name
        ]


def convert_to_numpy(
    ragged_or_dense: np.ndarray[Any, Any] | Sequence[Sequence[Any]] | Any,
    dtype: Any,
) -> np.ndarray[Any, Any]:
    """Converts a ragged or dense list of inputs to a ragged/dense numpy array.

    The output is adjusted to be 2D.

    Args:
        ragged_or_dense: Input that is either already a numpy array, or nested
            sequence.
        dtype: Numpy dtype of output array.

    Returns:
        Corresponding numpy array.
    """
    if hasattr(ragged_or_dense, "numpy"):
        # Support tf.RaggedTensor and other TF input dtypes.
        if callable(getattr(ragged_or_dense, "numpy")):
            ragged_or_dense = ragged_or_dense.numpy()

    if isinstance(ragged_or_dense, jax.Array):
        ragged_or_dense = np.asarray(ragged_or_dense)

    if isinstance(ragged_or_dense, np.ndarray):
        # Convert 1D to 2D.
        if ragged_or_dense.dtype != np.ndarray and ragged_or_dense.ndim == 1:
            return ragged_or_dense.reshape(-1, 1).astype(dtype)

        # If dense, return converted dense type.
        if ragged_or_dense.dtype != np.ndarray:
            return ragged_or_dense.astype(dtype)

        # Ragged numpy array.
        return ragged_or_dense

    # Handle 1D sequence input.
    if not isinstance(ragged_or_dense[0], collections.abc.Sequence):
        return np.asarray(ragged_or_dense, dtype=dtype).reshape(-1, 1)

    # Assemble elements into an nd-array.
    counts = [len(vals) for vals in ragged_or_dense]
    if all([count == counts[0] for count in counts]):
        # Dense input.
        return np.asarray(ragged_or_dense, dtype=dtype)
    else:
        # Ragged input, convert to ragged numpy arrays.
        return np.array(
            [np.array(row, dtype=dtype) for row in ragged_or_dense],
            dtype=np.ndarray,
        )


def ones_like(
    ragged_or_dense: np.ndarray[Any, Any], dtype: Any = None
) -> np.ndarray[Any, Any]:
    """Creates an array of ones the same as as the input.

    This differs from traditional numpy in that a ragged input will lead to
    a resulting ragged array of ones, whereas np.ones_like(...) will instead
    only consider the outer array and return a 1D dense array of ones.

    Args:
        ragged_or_dense: The ragged or dense input whose shape and data-type
              define these same attributes of the returned array.
        dtype: The data-type of the returned array.

    Returns:
        An array of ones with the same shape as the input, and specified data
        type.
    """
    dtype = dtype or ragged_or_dense.dtype
    if ragged_or_dense.dtype == np.ndarray:
        # Ragged.
        return np.array(
            [np.ones_like(row, dtype=dtype) for row in ragged_or_dense],
            dtype=np.ndarray,
        )
    else:
        # Dense.
        return np.ones_like(ragged_or_dense, dtype=dtype)


def create_feature_samples(
    feature_structure: Nested[T],
    feature_ids: Nested[ArrayLike | Sequence[int] | Sequence[Sequence[int]]],
    feature_weights: None
    | (Nested[ArrayLike | Sequence[float] | Sequence[Sequence[float]]]),
) -> Nested[FeatureSamples]:
    """Constructs a collection of sample tuples from provided IDs and weights.

    Args:
        feature_structure: The nested structure of the inputs (typically
            `FeatureSpec`s).
        feature_ids: The feature IDs to use for the samples.
        feature_weights: The feature weights to use for the samples.  Defaults
            to ones if not provided.

    Returns:
        A nested collection of `FeatureSamples` corresponding to the input IDs
        and weights, for use in embedding lookups.
    """
    # Create numpy arrays from inputs.
    feature_ids = jax.tree.map(
        lambda _, ids: convert_to_numpy(ids, np.int32),
        feature_structure,
        feature_ids,
    )

    if feature_weights is None:
        # Make ragged or dense ones_like.
        feature_weights = jax.tree.map(
            lambda _, ids: ones_like(ids, np.float32),
            feature_structure,
            feature_ids,
        )
    else:
        feature_weights = jax.tree.map(
            lambda _, wgts: convert_to_numpy(wgts, np.float32),
            feature_structure,
            feature_weights,
        )

    # Assemble.
    def _create_feature_samples(
        sample_ids: np.ndarray[Any, Any],
        sample_weights: np.ndarray[Any, Any],
    ) -> FeatureSamples:
        return FeatureSamples(sample_ids, sample_weights)

    output: Nested[FeatureSamples] = jax.tree.map(
        lambda _, sample_ids, sample_weights: _create_feature_samples(
            sample_ids, sample_weights
        ),
        feature_structure,
        feature_ids,
        feature_weights,
    )
    return output


def stack_and_shard_samples(
    feature_specs: Nested[FeatureSpec],
    feature_samples: Nested[FeatureSamples],
    local_device_count: int,
    global_device_count: int,
    num_sc_per_device: int,
    static_buffer_size: int | Mapping[str, int] | None = None,
) -> tuple[dict[str, ShardedCooMatrix], dict[str, InputStatsPerTable]]:
    """Prepares input samples for use in embedding lookups.

    Args:
        feature_specs: Nested collection of feature specifications.
        feature_samples: Nested collection of feature samples.
        local_device_count: Number of local JAX devices.
        global_device_count: Number of global JAX devices.
        num_sc_per_device: Number of sparsecores per device.
        static_buffer_size: The static buffer size to use for the samples.
          Defaults to None, in which case an upper-bound for the buffer size
          will be automatically determined.

    Returns:
        The preprocessed inputs, and statistics useful for updating FeatureSpecs
        based on the provided input data.
    """
    del static_buffer_size  # Currently ignored.
    flat_feature_specs, _ = jax.tree.flatten(feature_specs)

    feature_tokens = []
    feature_weights = []

    def collect_tokens_and_weights(
        feature_spec: FeatureSpec, samples: FeatureSamples
    ) -> None:
        del feature_spec
        feature_tokens.append(samples.tokens)
        feature_weights.append(samples.weights)

    jax.tree.map(collect_tokens_and_weights, feature_specs, feature_samples)

    preprocessed_inputs, stats = embedding.preprocess_sparse_dense_matmul_input(
        feature_tokens,
        feature_weights,
        flat_feature_specs,
        local_device_count=local_device_count,
        global_device_count=global_device_count,
        num_sc_per_device=num_sc_per_device,
        sharding_strategy="MOD",
        has_leading_dimension=False,
        allow_id_dropping=True,
    )

    out: dict[str, ShardedCooMatrix] = {}
    out_stats: dict[str, InputStatsPerTable] = {}
    tables_names = preprocessed_inputs.lhs_row_pointers.keys()
    for table_name in tables_names:
        shard_ends = preprocessed_inputs.lhs_row_pointers[table_name]
        shard_starts = np.concatenate(
            [np.asarray([0]), _round_up_to_multiple(shard_ends[:-1], 8)]
        )
        out[table_name] = ShardedCooMatrix(
            shard_starts=shard_starts,
            shard_ends=shard_ends,
            col_ids=preprocessed_inputs.lhs_embedding_ids[table_name],
            row_ids=preprocessed_inputs.lhs_sample_ids[table_name],
            values=preprocessed_inputs.lhs_gains[table_name],
        )
        out_stats[table_name] = InputStatsPerTable(
            max_ids_per_partition=np.max(
                stats.max_ids_per_partition[table_name]
            ),
            max_unique_ids_per_partition=np.max(
                stats.max_unique_ids_per_partition[table_name]
            ),
            required_buffer_size_per_device=np.max(
                stats.required_buffer_size_per_sc[table_name]
            )
            * num_sc_per_device,
        )

    return out, out_stats



"""JAX implementation of the TPU embedding layer."""

import math
import typing
from typing import Any, Mapping, Sequence, Union

import jax
import keras
import numpy as np
from jax import numpy as jnp
from jax.experimental import layout as jax_layout
from jax_tpu_embedding.sparsecore.lib.nn import embedding
from jax_tpu_embedding.sparsecore.lib.nn import embedding_spec
from jax_tpu_embedding.sparsecore.lib.nn import (
    table_stacking as jte_table_stacking,
)
from jax_tpu_embedding.sparsecore.utils import utils as jte_utils

from keras_rs.src import types
from keras_rs.src.layers.embedding import base_distributed_embedding
from keras_rs.src.layers.embedding import distributed_embedding_config as config
from keras_rs.src.layers.embedding.jax import config_conversion
from keras_rs.src.layers.embedding.jax import (
    embedding_lookup as jte_embedding_lookup,
)
from keras_rs.src.layers.embedding.jax import embedding_utils
from keras_rs.src.types import Nested
from keras_rs.src.utils import keras_utils

ArrayLike = Union[np.ndarray[Any, Any], jax.Array]
FeatureConfig = config.FeatureConfig
shard_map = jax.experimental.shard_map.shard_map  # type: ignore[attr-defined]


def _get_partition_spec(
    layout: (
        keras.distribution.TensorLayout
        | jax_layout.Format
        | jax.sharding.NamedSharding
        | jax.sharding.PartitionSpec
    ),
) -> Any:
    """Extracts the partition spec from a layout or sharding."""
    if isinstance(layout, keras.distribution.TensorLayout):
        layout = layout.backend_layout

    if isinstance(layout, jax_layout.Format):
        layout = layout.sharding

    if isinstance(layout, jax.sharding.NamedSharding):
        layout = layout.spec

    return layout


class ShardedInitializer(keras.initializers.Initializer):
    """Wraps an initializer to prepare for use with embedding tables.

    Jit-compiles the function and applies optimal output sharding to
    allow initialization on device.
    """

    def __init__(
        self,
        initializer: keras.initializers.Initializer | str,
        layout: keras.distribution.TensorLayout | None,
    ):
        if isinstance(initializer, str):
            initializer = keras.initializers.get(initializer)

        self._initializer = initializer
        self._layout = layout

    def __call__(
        self, shape: types.Shape, dtype: types.DType | None = None
    ) -> jax.Array:
        if self._layout is not None:
            compiled_initializer = jax.jit(
                self._initializer,
                out_shardings=self._layout.backend_layout,
                static_argnames=["shape", "dtype"],
            )
            output: jax.Array = compiled_initializer(shape, dtype)
            return output

        output = self._initializer(shape, dtype)
        return output


class StackedTableInitializer(keras.initializers.Initializer):
    """Initializes a single stacked table from multiple table initializers."""

    def __init__(
        self,
        table_specs: Nested[embedding_spec.TableSpec],
        num_shards: int,
        layout: keras.distribution.TensorLayout,
        seed: int | keras.random.SeedGenerator | jax.Array = 0,
    ):
        # Sort table specs so we can simply concatenate them when assembling the
        # stacked table.
        self._table_specs = sorted(
            keras.tree.flatten(table_specs),
            key=lambda table_spec: (
                table_spec.setting_in_stack.row_offset_in_shard,
            ),
        )
        self._num_shards = num_shards
        self._layout = layout
        self._key = keras.src.backend.jax.random.jax_draw_seed(seed)

    def _initialize_shard(
        self,
        keys: jax.Array,
        shape: tuple[int, int],
        dtype: Any,
        num_shards_per_device: int,
    ) -> jax.Array:
        """Initializes a single shard of a stacked table."""
        del shape  # Unused.
        table_shards: list[jax.Array] = []
        # NOTE: the following ignores padding, rotations in shard, and
        # mod-sharding, assuming all initializers are shard-independent.
        for i in range(num_shards_per_device):
            for j, table_spec in enumerate(self._table_specs):
                setting_in_stack = table_spec.setting_in_stack
                table_shard_shape = (
                    setting_in_stack.padded_vocab_size // self._num_shards,
                    setting_in_stack.padded_embedding_dim,
                )
                initializer = table_spec.initializer
                table_shards.append(
                    initializer(keys[i, j], table_shard_shape, dtype)
                )

        return jnp.concatenate(table_shards, axis=0)

    def __call__(
        self, shape: types.Shape, dtype: types.DType | None = None
    ) -> jax.Array:
        stacked_table_spec = typing.cast(
            embedding_spec.StackedTableSpec,
            self._table_specs[0].stacked_table_spec,
        )

        # Input shape is governed by the table specs.
        assert shape == (
            stacked_table_spec.stack_vocab_size,
            stacked_table_spec.stack_embedding_dim,
        )

        layout = self._layout
        backend_layout = layout.backend_layout
        backend_mesh = layout.device_mesh.backend_mesh
        num_devices_along_axis = backend_mesh.shape[layout.axes[0]]
        num_shards_per_device = self._num_shards // num_devices_along_axis
        shard_shape = (
            stacked_table_spec.stack_vocab_size // num_devices_along_axis,
            stacked_table_spec.stack_embedding_dim,
        )

        sharded_initializer = jax.jit(
            shard_map(
                lambda keys: self._initialize_shard(
                    keys, shard_shape, dtype, num_shards_per_device
                ),
                mesh=backend_mesh,
                in_specs=_get_partition_spec(backend_layout),
                out_specs=_get_partition_spec(backend_layout),
            ),
            out_shardings=backend_layout,
        )

        keys = jax.random.split(
            self._key, (self._num_shards, len(self._table_specs))
        )
        # Try extracting seeds from the existing table initializers.
        for i, table_spec in enumerate(self._table_specs):
            initializer = table_spec.initializer
            if isinstance(
                initializer, config_conversion.WrappedKerasInitializer
            ):
                initializer_key = initializer.key()
                if initializer_key is not None:
                    col = jax.random.split(initializer_key, self._num_shards)
                    keys = keys.at[:, i].set(col)

        output: jax.Array = sharded_initializer(keys)
        return output


class DistributedEmbedding(base_distributed_embedding.DistributedEmbedding):
    """JAX implementation of the TPU embedding layer."""

    def _create_sparsecore_distribution(
        self, sparsecore_axis_name: str = "sparsecore"
    ) -> tuple[
        keras.distribution.ModelParallel, keras.distribution.TensorLayout
    ]:
        """SparseCore requires a specific layout.

        The mesh must be 1D, must use all TPUs available, and must shard all
        tables across all devices.

        Args:
            sparsecore_axis_name: The name of the sparsecore axis.

        Returns:
            A Keras distribution to use for all sparsecore operations.
        """
        all_devices = jax.devices()
        axes = [sparsecore_axis_name]
        device_mesh = keras.distribution.DeviceMesh(
            (len(all_devices),), axes, all_devices
        )
        sparsecore_layout = keras.distribution.TensorLayout(axes, device_mesh)
        # Custom sparsecore layout with tiling.
        LayoutClass = (
            jax_layout.Layout
            if jax.__version_info__ >= (0, 6, 3)
            else jax_layout.DeviceLocalLayout  # type: ignore
        )
        layout = (
            LayoutClass(major_to_minor=(0, 1), tiling=((8,),))  # type: ignore
            if jax.__version_info__ >= (0, 7, 1)
            else LayoutClass(major_to_minor=(0, 1), _tiling=((8,),))  # type: ignore
        )
        # pylint: disable-next=protected-access
        sparsecore_layout._backend_layout = jax_layout.Format(
            layout,  # type: ignore
            jax.sharding.NamedSharding(
                device_mesh.backend_mesh,
                jax.sharding.PartitionSpec(
                    axes  # type: ignore[no-untyped-call]
                ),
            ),
        )
        layout_map = keras.distribution.LayoutMap(device_mesh=device_mesh)
        path = self.path
        if path is None:
            # Layer hasn't been properly built yet.  Use current layer name.
            path = self.name
        layout_map[path + "/var"] = sparsecore_layout
        sparsecore_distribution = keras.distribution.ModelParallel(
            layout_map=layout_map
        )
        return sparsecore_distribution, sparsecore_layout

    def _add_sparsecore_weight(
        self,
        name: str,
        shape: tuple[int, int],
        initializer: jax.nn.initializers.Initializer,
        dtype: Any,
        overwrite_with_gradient: bool,
    ) -> keras.Variable:
        var = self.add_weight(
            name=name, shape=shape, initializer=initializer, dtype=dtype
        )
        var.overwrite_with_gradient = overwrite_with_gradient
        return var

    def _add_table_variable(
        self,
        table_specs: Sequence[embedding_spec.TableSpec],
        num_shards: int,
        add_slot_variables: bool,
    ) -> tuple[keras.Variable, tuple[keras.Variable, ...] | None]:
        stacked_table_spec = typing.cast(
            embedding_spec.StackedTableSpec, table_specs[0].stacked_table_spec
        )
        optimizer = stacked_table_spec.optimizer
        num_slot_variables = optimizer.slot_variables_count()
        table_shape = (
            stacked_table_spec.stack_vocab_size,
            stacked_table_spec.stack_embedding_dim,
        )

        # Make a stacked embedding table initializer.
        table_initializers = [
            config_conversion.jax_to_keras_initializer(table_spec.initializer)
            for table_spec in table_specs
        ]
        # If all initializers are the same, we can use a single sharded
        # initializer. Otherwise, we need to interleave individual stacked table
        # shards.
        sparsecore_layout = self._sparsecore_layout
        stacked_table_initializer = ShardedInitializer(
            table_initializers[0], sparsecore_layout
        )
        if not all(
            initializer == table_initializers[0]
            for initializer in table_initializers
        ):
            stacked_table_initializer = StackedTableInitializer(
                table_specs, num_shards, sparsecore_layout
            )

        variable_name = f"var:{stacked_table_spec.stack_name}:table"
        table_variable = self._add_sparsecore_weight(
            name=variable_name,
            shape=table_shape,
            initializer=stacked_table_initializer,
            dtype="float32",
            overwrite_with_gradient=True,
        )

        slot_variables = None
        if add_slot_variables:
            # All optimizers for a given stacked table are guaranteed to be the
            # same, so we can use a single sharded initializer for the entire
            # stacked table.
            slot_initializers = optimizer.slot_variables_initializers()
            # Try extracting field names from variables, otherwise just use the
            # count.
            slot_names = range(num_slot_variables)
            if hasattr(slot_initializers, "_fields"):
                slot_names = slot_initializers._fields

            slot_variables = tuple(
                self._add_sparsecore_weight(
                    name=f"{variable_name}:slot:{slot_name}",
                    shape=table_shape,
                    initializer=ShardedInitializer(
                        config_conversion.jax_to_keras_initializer(initializer),
                        sparsecore_layout,
                    ),
                    dtype=jnp.float32,
                    overwrite_with_gradient=True,
                )
                for slot_name, initializer in zip(slot_names, slot_initializers)
            )
            slot_variables = keras.tree.pack_sequence_as(
                slot_initializers, slot_variables
            )

        return table_variable, slot_variables

    @keras_utils.no_automatic_dependency_tracking
    def _sparsecore_init(
        self,
        feature_configs: dict[str, FeatureConfig],
        table_stacking: str | Sequence[str] | Sequence[Sequence[str]],
    ) -> None:
        if not self._has_sparsecore():
            raise ValueError(
                "Not sparse cores available, cannot use explicit sparsecore"
                " placement."
            )

        self._sc_feature_configs = feature_configs
        self._sparsecore_built = False
        # Fill in any empty default settings.
        for feature_config in keras.tree.flatten(self._sc_feature_configs):
            if feature_config.table.initializer is None:
                table = feature_config.table
                table.initializer = keras.initializers.TruncatedNormal(
                    mean=0.0, stddev=1.0 / math.sqrt(float(table.embedding_dim))
                )

        # Actual stacking of tables is done in build() to ensure the
        # distribution is set up correctly.
        self._table_stacking = table_stacking

    def _sparsecore_build(
        self, input_shapes: Nested[types.Shape] | None = None
    ) -> None:
        self.sparsecore_build(input_shapes)

    @keras_utils.no_automatic_dependency_tracking
    def sparsecore_build(
        self, input_shapes: Nested[types.Shape] | None = None
    ) -> None:
        del input_shapes  # Unused.

        if self._sparsecore_built:
            return

        feature_specs = config_conversion.keras_to_jte_feature_configs(
            self._sc_feature_configs
        )

        # Distribution for sparsecore operations.
        sparsecore_distribution, sparsecore_layout = (
            self._create_sparsecore_distribution()
        )
        self._sparsecore_layout = sparsecore_layout
        self._sparsecore_distribution = sparsecore_distribution

        mesh = sparsecore_distribution.device_mesh.backend_mesh
        global_device_count = mesh.devices.size
        num_sc_per_device = jte_utils.num_sparsecores_per_device(
            mesh.devices.item(0)
        )
        # One table shard per global sparsecore.
        num_variable_shards = global_device_count * num_sc_per_device

        # Maybe stack tables.
        table_stacking = self._table_stacking
        if table_stacking is not None:
            if isinstance(table_stacking, str):
                if table_stacking == "auto":
                    jte_table_stacking.auto_stack_tables(
                        feature_specs, global_device_count, num_sc_per_device
                    )
                else:
                    raise ValueError(
                        f"Unsupported table stacking {table_stacking}, must be"
                        "None, 'auto', or sequences of table names to stack."
                    )
            else:
                if isinstance(table_stacking, list) and len(table_stacking) > 0:
                    elem = table_stacking[0]
                    # List of lists of table names.
                    if isinstance(elem, list):
                        for table_names in table_stacking:
                            jte_table_stacking.stack_tables(
                                feature_specs,
                                table_names,
                                global_device_count,
                                num_sc_per_device,
                            )
                    # Single list of table names.
                    elif isinstance(elem, str):
                        jte_table_stacking.stack_tables(
                            feature_specs,
                            table_stacking,
                            global_device_count,
                            num_sc_per_device,
                        )
                    else:
                        raise ValueError(
                            f"Unsupported table stacking {table_stacking}, "
                            "must be None, 'auto', or sequences of table names "
                            "to stack."
                        )

        # Adjust any non-stacked tables to prepare for training.
        embedding.prepare_feature_specs_for_training(
            feature_specs, global_device_count, num_sc_per_device
        )

        # Collect all stacked tables.
        table_specs = embedding_utils.get_table_specs(feature_specs)
        table_stacks = embedding_utils.get_table_stacks(table_specs)

        # Create variables for all stacked tables and slot variables.
        with sparsecore_distribution.scope():
            self._table_and_slot_variables = {
                table_name: self._add_table_variable(
                    table_stack,
                    add_slot_variables=self.trainable,
                    num_shards=num_variable_shards,
                )
                for table_name, table_stack in table_stacks.items()
            }

            # Create a step-counter variable for use in custom table gradients.
            # This must be a floating-point type so we can get a real gradient
            # for it. It will automatically be updated with each application of
            # the optimizer, since the next iteration is returned in the
            # gradient.
            sharded_zero_initializer = ShardedInitializer(
                "zeros",
                keras.distribution.TensorLayout(
                    [], sparsecore_layout.device_mesh
                ),
            )
            self._iterations = self.add_weight(
                shape=(),
                name="iteration",
                initializer=sharded_zero_initializer,
                dtype="float32",
                trainable=True,
            )
            self._iterations.overwrite_with_gradient = True

        self._config = jte_embedding_lookup.EmbeddingLookupConfiguration(
            feature_specs,
            mesh=mesh,
            table_partition=_get_partition_spec(sparsecore_layout),
            samples_partition=_get_partition_spec(sparsecore_layout),
            table_layout=sparsecore_layout.backend_layout,
        )

        self._sparsecore_built = True

    def _sparsecore_symbolic_preprocess(
        self,
        inputs: dict[str, types.Tensor],
        weights: dict[str, types.Tensor] | None,
        training: bool = False,
    ) -> dict[str, dict[str, embedding_utils.ShardedCooMatrix]]:
        """Allow preprocess(...) with `keras.Input`s.

        This is to support creating functional models via:
        ```python
        inputs = keras.Input(shape=(None), dtype="int32")
        weights = keras.Input(shape=(None), dtype="float32")
        preprocessed_inputs = distributed_embedding.preprocess(inputs, weights)
        outputs = distributed_embedding(preprocessed_inputs)
        model = keras.Model(inputs=preprocessed_inputs, outputs=outputs)
        ```

        Args:
            inputs: SparseCore path->tensor input ID's tensors.
            weights: Optional Sparsecore path->tensor input weights tensors.
            training: Whether the layer is training or not.

        Returns:
            Symbolic preprocessed input tensors to the layer/model.
        """
        # Arguments are currently ignored since the input shape is governed
        # by the stacked table configuration.
        del inputs, weights, training

        # Each stacked-table gets a ShardedCooMatrix.
        table_specs = embedding_utils.get_table_specs(
            self._config.feature_specs
        )
        table_stacks = embedding_utils.get_table_stacks(table_specs)
        stacked_table_specs = {
            stack_name: stack[0].stacked_table_spec
            for stack_name, stack in table_stacks.items()
        }

        def _compute_table_output_spec(
            stacked_table_spec: embedding_spec.StackedTableSpec,
        ) -> embedding_utils.ShardedCooMatrix:
            # The true shape of the components in the ShardedCooMatrix depends
            # on the hardware configuration (# devices, sparsecores),
            # properties of the input data (# max IDs, unique IDs), and other
            # hints like a suggested internal buffer size.  Some of the
            # calculations are currently a bit in flux as we experiment with
            # memory trade-offs.  For the purposes of input/output sizes,
            # however, the size could be viewed as dynamic 1D without affecting
            # the output spec sizes.
            del stacked_table_spec
            return embedding_utils.ShardedCooMatrix(
                # Mark these as `Input`s since that's how they will be used when
                # constructing a functional Keras model.
                shard_starts=keras.Input(shape=tuple(), dtype="int32"),
                shard_ends=keras.Input(shape=tuple(), dtype="int32"),
                col_ids=keras.Input(shape=tuple(), dtype="int32"),
                row_ids=keras.Input(shape=tuple(), dtype="int32"),
                values=keras.Input(shape=tuple(), dtype="float32"),
            )

        preprocessed = keras.tree.map_structure(
            _compute_table_output_spec, stacked_table_specs
        )

        return {"inputs": preprocessed}

    def _sparsecore_preprocess(
        self,
        inputs: dict[str, types.Tensor],
        weights: dict[str, types.Tensor] | None,
        training: bool = False,
    ) -> dict[str, dict[str, embedding_utils.ShardedCooMatrix]]:
        if any(
            isinstance(x, jax.core.Tracer) for x in keras.tree.flatten(inputs)
        ):
            raise ValueError(
                "DistributedEmbedding.preprocess(...) does not support"
                " jit-compilation"
            )

        if not self._sparsecore_built:
            self._sparsecore_build()

        # Support symbolic KerasTensors (i.e. keras.Input).
        if any(
            isinstance(x, keras.KerasTensor) for x in keras.tree.flatten(inputs)
        ):
            return self._sparsecore_symbolic_preprocess(
                inputs, weights, training
            )

        samples = embedding_utils.create_feature_samples(
            self._config.feature_specs, inputs, weights
        )

        layout = self._sparsecore_layout
        mesh = layout.device_mesh.backend_mesh
        global_device_count = mesh.devices.size
        local_device_count = mesh.local_mesh.devices.size
        num_sc_per_device = jte_utils.num_sparsecores_per_device(
            mesh.devices.item(0)
        )

        preprocessed, stats = embedding_utils.stack_and_shard_samples(
            self._config.feature_specs,
            samples,
            local_device_count,
            global_device_count,
            num_sc_per_device,
        )

        if training:
            # Synchronize input statistics across all devices and update the
            # underlying stacked tables specs in the feature specs.
            prev_stats = embedding_utils.get_stacked_table_stats(
                self._config.feature_specs
            )

            # Take the maximum with existing stats.
            stats = keras.tree.map_structure(max, prev_stats, stats)

            # Flatten the stats so we can more efficiently transfer them
            # between hosts.  We use jax.tree because we will later need to
            # unflatten.
            flat_stats, stats_treedef = jax.tree.flatten(stats)

            # In the case of multiple local CPU devices per host, we need to
            # replicate the stats to placate JAX collectives.
            num_local_cpu_devices = jax.local_device_count("cpu")
            tiled_stats = np.tile(
                np.array(flat_stats, dtype=np.int32), (num_local_cpu_devices, 1)
            )

            # Aggregate variables across all processes/devices.
            max_across_cpus = jax.pmap(
                lambda x: jax.lax.pmax(  # type: ignore[no-untyped-call]
                    x, "all_cpus"
                ),
                axis_name="all_cpus",
                backend="cpu",
            )
            flat_stats = max_across_cpus(tiled_stats)[0].tolist()
            stats = jax.tree.unflatten(stats_treedef, flat_stats)

            # Update configuration and repeat preprocessing if stats changed.
            if stats != prev_stats:
                embedding_utils.update_stacked_table_stats(
                    self._config.feature_specs, stats
                )

                # Re-execute preprocessing with consistent input statistics.
                preprocessed, _ = embedding_utils.stack_and_shard_samples(
                    self._config.feature_specs,
                    samples,
                    local_device_count,
                    global_device_count,
                    num_sc_per_device,
                )

        return {"inputs": preprocessed}

    def _sparsecore_call(
        self,
        inputs: dict[str, types.Tensor],
        weights: dict[str, types.Tensor] | None = None,
        training: bool = False,
        **kwargs: Any,
    ) -> dict[str, types.Tensor]:
        assert weights is None

        if not self._sparsecore_built:
            self._sparsecore_build()

        table_and_slots = keras.tree.map_structure(
            lambda var: var.value, self._table_and_slot_variables
        )
        with self._sparsecore_distribution.scope():
            lookup_func = jax.jit(
                jte_embedding_lookup.embedding_lookup, static_argnames="config"
            )
            out: dict[str, types.Tensor] = lookup_func(
                self._config, inputs, table_and_slots, self._iterations.value
            )
            return out

    def set_embedding_tables(self, tables: Mapping[str, ArrayLike]) -> None:
        """Sets the embedding tables to specific (unsharded) values.

        Args:
          tables: Mapping of table name -> table values.
        """
        if "default_device" in self._placement_to_path_to_feature_config:
            self._default_device_set_tables(tables)

        if "sparsecore" in self._placement_to_path_to_feature_config:
            self._sparsecore_set_tables(tables)

    def _default_device_set_tables(
        self, tables: Mapping[str, ArrayLike]
    ) -> None:
        if not self.built:
            raise ValueError("Layer must first be built before setting tables.")

        if "default_device" in self._placement_to_path_to_feature_config:
            table_name_to_embedding_layer = {}
            for (
                path,
                feature_config,
            ) in self._placement_to_path_to_feature_config[
                "default_device"
            ].items():
                table_name_to_embedding_layer[feature_config.table.name] = (
                    self._default_device_embedding_layers[path]
                )

            for (
                table_name,
                embedding_layer,
            ) in table_name_to_embedding_layer.items():
                table_values = tables.get(table_name, None)
                if table_values is not None:
                    if embedding_layer.lora_enabled:
                        raise ValueError("Cannot set table if LoRA is enabled.")
                    # pylint: disable-next=protected-access
                    embedding_layer._embeddings.assign(table_values)

    def _sparsecore_set_tables(self, tables: Mapping[str, ArrayLike]) -> None:
        if not self._sparsecore_built:
            self._sparsecore_build()

        config = self._config
        num_table_shards = config.mesh.devices.size * config.num_sc_per_device
        table_specs = embedding_utils.get_table_specs(config.feature_specs)
        sharded_tables = embedding_utils.stack_and_shard_tables(
            table_specs,
            tables,
            num_table_shards,
        )

        device_tables = jax.device_put(
            jax.tree.map(
                # Flatten shard dimension to allow auto-sharding to split the
                # array.
                lambda table: table.reshape((-1, table.shape[-1])),
                sharded_tables,
            ),
            self._sparsecore_layout.backend_layout,
        )

        # Assign stacked table variables to the device values.
        keras.tree.map_structure_up_to(
            device_tables,
            lambda table_and_slot_variables,
            table_value: table_and_slot_variables[0].assign(table_value),
            self._table_and_slot_variables,
            device_tables,
        )

    def _sparsecore_get_embedding_tables(self) -> dict[str, ArrayLike]:
        if not self._sparsecore_built:
            self.sparsecore_build()

        config = self._config
        num_table_shards = config.mesh.devices.size * config.num_sc_per_device
        table_specs = embedding_utils.get_table_specs(config.feature_specs)

        # Extract only the table variables, not the gradient slot variables.
        table_variables = {
            name: jax.device_get(table_and_slots[0].value)
            for name, table_and_slots in self._table_and_slot_variables.items()
        }

        return typing.cast(
            dict[str, ArrayLike],
            embedding_utils.unshard_and_unstack_tables(
                table_specs, table_variables, num_table_shards
            ),
        )



"""Conversion utilities for Keras DistributedEmbeddingConfig to JAX."""

import inspect
import math
import random as python_random
from typing import Any, Callable

import jax
import jax.numpy as jnp
import keras
import numpy as np
from jax_tpu_embedding.sparsecore.lib.nn import embedding_spec

from keras_rs.src import types
from keras_rs.src.layers.embedding import distributed_embedding_config as config
from keras_rs.src.types import Nested


class WrappedKerasInitializer(jax.nn.initializers.Initializer):
    """Wraps a Keras initializer for use in JAX."""

    def __init__(self, initializer: keras.initializers.Initializer):
        if isinstance(initializer, str):
            initializer = keras.initializers.get(initializer)
        self.initializer = initializer

    def key(self) -> jax.Array | None:
        """Extract a key from the underlying keras initializer."""
        # All built-in keras initializers have a `seed` attribute.
        # Extract this and turn it into a key for use with JAX.
        if hasattr(self.initializer, "seed"):
            output: jax.Array = keras.src.backend.jax.random.jax_draw_seed(
                self.initializer.seed
            )
            return output
        return None

    def __call__(
        self,
        key: Any,
        shape: Any,
        dtype: Any = jnp.float_,
        out_sharding: Any = None,
    ) -> jax.Array:
        del out_sharding
        # Force use of provided key.  The JAX backend for random initializers
        # forwards the `seed` attribute to the underlying JAX random functions.
        if key is not None and hasattr(self.initializer, "seed"):
            old_seed = self.initializer.seed
            self.initializer.seed = key
            out: jax.Array = self.initializer(shape, dtype)
            self.initializer.seed = old_seed
            return out

        output: jax.Array = self.initializer(shape, dtype)
        return output


# pylint: disable-next=g-classes-have-attributes
class WrappedJaxInitializer(keras.initializers.Initializer):
    """Wraps a JAX initializer for use in Keras.

    Attributes:
        initializer: The wrapped JAX initializer.

    Args:
        initializer: The JAX initializer to wrap.
        seed: Optional Keras seed for use with random JAX initializers.
    """

    def __init__(
        self,
        initializer: jax.nn.initializers.Initializer,
        seed: int | keras.random.SeedGenerator | None = None,
    ):
        self.initializer = initializer
        if seed is None:
            # Consistency with keras.random.make_default_seed().
            seed = python_random.randint(1, int(1e9))
        self.seed = seed

    def key(self) -> jax.Array:
        """Converts the interal seed to a JAX random key."""
        seed = self.seed
        if isinstance(seed, int):
            return jax.random.key(self.seed)
        elif isinstance(seed, keras.random.SeedGenerator):
            return jax.random.key(seed.next())
        elif isinstance(seed, jax.Array):
            return seed
        else:
            raise ValueError(f"Unknown seed {seed} of type {type(seed)}.")

    def __call__(
        self,
        shape: types.Shape,
        dtype: types.DType | None = None,
        **kwargs: Any,
    ) -> jax.Array:
        del kwargs  # Unused.
        return self.initializer(self.key(), shape, dtype)


def keras_to_jax_initializer(
    initializer: str | keras.initializers.Initializer,
) -> jax.nn.initializers.Initializer:
    """Converts a Keras initializer to a JAX initializer.

    Args:
        initializer: Keras initializer to convert.

    Returns:
        A JAX-compatible equivalent initializer.
    """
    if isinstance(initializer, WrappedJaxInitializer):
        return initializer.initializer
    return WrappedKerasInitializer(initializer)


def jax_to_keras_initializer(
    initializer: jax.nn.initializers.Initializer,
) -> keras.initializers.Initializer:
    """Converts a JAX initializer to a Keras initializer.

    Args:
        initializer: JAX initializer to convert.

    Returns:
        An equivalent Keras initializer.
    """
    if isinstance(initializer, WrappedKerasInitializer):
        return initializer.initializer
    return WrappedJaxInitializer(initializer)


def keras_to_jte_learning_rate(
    learning_rate: keras.Variable | float | Callable[..., float],
) -> float | Callable[..., float]:
    """Converts a Keras learning rate to a JAX TPU Embedding learning rate.

    Args:
      learning_rate: Any Keras-compatible learning-rate type.  If a Callable,
          must either take no parameters, or the step size as a single argument.

    Returns:
        A JAX TPU Embedding learning rate.

    Raises:
        ValueError if the learning rate is not supported.
    """

    # Supported keras optimizer general options.
    if isinstance(learning_rate, keras.Variable):
        # Extract the first (and only) element of the variable.
        learning_rate = np.array(learning_rate.value, dtype=float)
        assert learning_rate.size == 1
        lr_float: float = learning_rate.item(0)
        return lr_float
    elif callable(learning_rate):
        # Callable learning rate functions are expected to take a singular step
        # count argument, or no arguments.
        args = inspect.getfullargspec(learning_rate).args
        # If not a function, then it's an object instance with `self` as the
        # first argument.
        num_args = (
            len(args) if inspect.isfunction(learning_rate) else len(args) - 1
        )
        if num_args <= 1:
            return learning_rate
    elif isinstance(learning_rate, float):
        return learning_rate

    raise ValueError(
        f"Unsupported learning rate: {learning_rate} of type"
        f" {type(learning_rate)}."
    )


def jte_to_keras_learning_rate(
    learning_rate: float | Callable[..., float],
) -> float | Callable[..., float]:
    """Converts a JAX TPU Embedding learning rate to a Keras learning rate.

    Args:
        learning_rate: The learning rate value or function.  If a Callable, must
            either take no parameters, or the step size as a single argument.

    Returns:
        A JAX TPU Embedding learning rate.

    Raises:
        ValueError if the learning rate is not supported.
    """
    if callable(learning_rate):
        # Callable learning rate functions are expected to take a singular step
        # count argument, or no arguments.
        args = inspect.getfullargspec(learning_rate).args
        # If not a function, then it's an object instance, with `self` as the
        # first arguments.
        num_args = (
            len(args) if inspect.isfunction(learning_rate) else len(args) - 1
        )
        if num_args <= 1:
            return learning_rate
    elif isinstance(learning_rate, float):
        return learning_rate

    raise ValueError(f"Unknown learning rate {learning_rate}")


def keras_to_jte_optimizer(
    optimizer: keras.optimizers.Optimizer | str,
) -> embedding_spec.OptimizerSpec:
    """Converts a Keras optimizer to a JAX TPU Embedding optimizer.

    Args:
        optimizer: Any Keras-compatible optimizer.

    Returns:
        A JAX TPU Embedding optimizer.
    """
    if isinstance(optimizer, str):
        optimizer = keras.optimizers.get(optimizer)

    # We need to extract the actual internal learning_rate function.
    # Unfortunately, the optimizer.learning_rate attribute tries to be smart,
    # and evaluates the learning rate at the current iteration step, which is
    # not what we want.
    # pylint: disable-next=protected-access
    learning_rate = keras_to_jte_learning_rate(optimizer._learning_rate)

    # Unsupported keras optimizer general options.
    if optimizer.clipnorm is not None:
        raise ValueError("Unsupported optimizer option `clipnorm`.")
    if optimizer.global_clipnorm is not None:
        raise ValueError("Unsupported optimizer option `global_clipnorm`.")
    if optimizer.use_ema:
        raise ValueError("Unsupported optimizer option `use_ema`.")
    if optimizer.loss_scale_factor is not None:
        raise ValueError("Unsupported optimizer option `loss_scale_factor`.")

    # Supported optimizers.
    if isinstance(optimizer, keras.optimizers.SGD):
        if getattr(optimizer, "nesterov", False):
            raise ValueError("Unsupported optimizer option `nesterov`.")
        if getattr(optimizer, "momentum", 0.0) != 0.0:
            raise ValueError("Unsupported optimizer option `momentum`.")
        return embedding_spec.SGDOptimizerSpec(learning_rate=learning_rate)
    elif isinstance(optimizer, keras.optimizers.Adagrad):
        if getattr(optimizer, "epsilon", 1e-7) != 1e-7:
            raise ValueError("Unsupported optimizer option `epsilon`.")
        return embedding_spec.AdagradOptimizerSpec(
            learning_rate=learning_rate,
            initial_accumulator_value=optimizer.initial_accumulator_value,
        )
    elif isinstance(optimizer, keras.optimizers.Adam):
        if getattr(optimizer, "amsgrad", False):
            raise ValueError("Unsupported optimizer option `amsgrad`.")

        return embedding_spec.AdamOptimizerSpec(
            learning_rate=learning_rate,
            beta_1=optimizer.beta_1,
            beta_2=optimizer.beta_2,
            epsilon=optimizer.epsilon,
        )
    elif isinstance(optimizer, keras.optimizers.Ftrl):
        if (
            getattr(optimizer, "l2_shrinkage_regularization_strength", 0.0)
            != 0.0
        ):
            raise ValueError(
                "Unsupported optimizer option "
                "`l2_shrinkage_regularization_strength`."
            )

        return embedding_spec.FTRLOptimizerSpec(
            learning_rate=learning_rate,
            learning_rate_power=optimizer.learning_rate_power,
            l1_regularization_strength=optimizer.l1_regularization_strength,
            l2_regularization_strength=optimizer.l2_regularization_strength,
            beta=optimizer.beta,
            initial_accumulator_value=optimizer.initial_accumulator_value,
        )

    raise ValueError(
        f"Unsupported optimizer type {type(optimizer)}. Optimizer must be "
        f"one of [Adagrad, Adam, Ftrl, SGD]."
    )


def jte_to_keras_optimizer(
    optimizer: embedding_spec.OptimizerSpec,
) -> keras.optimizers.Optimizer:
    """Converts a JAX TPU Embedding optimizer to a Keras optimizer.

    Args:
        optimizer: The JAX TPU Embedding optimizer.

    Returns:
        A corresponding Keras optimizer.
    """
    learning_rate = jte_to_keras_learning_rate(optimizer.learning_rate)
    if isinstance(optimizer, embedding_spec.SGDOptimizerSpec):
        return keras.optimizers.SGD(learning_rate=learning_rate)
    elif isinstance(optimizer, embedding_spec.AdagradOptimizerSpec):
        return keras.optimizers.Adagrad(
            learning_rate=learning_rate,
            initial_accumulator_value=optimizer.initial_accumulator_value,
        )
    elif isinstance(optimizer, embedding_spec.AdamOptimizerSpec):
        return keras.optimizers.Adam(
            learning_rate=learning_rate,
            beta_1=optimizer.beta_1,
            beta_2=optimizer.beta_2,
            epsilon=optimizer.epsilon,
        )
    elif isinstance(optimizer, embedding_spec.FTRLOptimizerSpec):
        if getattr(optimizer, "initial_linear_value", 0.0) != 0.0:
            raise ValueError(
                "Unsupported optimizer option `initial_linear_value`."
            )
        if getattr(optimizer, "multiply_linear_by_learning_rate", False):
            raise ValueError(
                "Unsupported optimizer option "
                "`multiply_linear_by_learning_rate`."
            )
        return keras.optimizers.Ftrl(
            learning_rate=learning_rate,
            learning_rate_power=optimizer.learning_rate_power,
            initial_accumulator_value=optimizer.initial_accumulator_value,
            l1_regularization_strength=optimizer.l1_regularization_strength,
            l2_regularization_strength=optimizer.l2_regularization_strength,
            beta=optimizer.beta,
        )

    raise ValueError(f"Unknown optimizer spec {type(optimizer)}.")


def _keras_to_jte_table_config(
    table_config: config.TableConfig,
) -> embedding_spec.TableSpec:
    # Initializer could be none.  Default to truncated normal.
    initializer = table_config.initializer
    if initializer is None:
        initializer = keras.initializers.TruncatedNormal(
            mean=0.0, stddev=1.0 / math.sqrt(float(table_config.embedding_dim))
        )
    return embedding_spec.TableSpec(
        name=table_config.name,
        vocabulary_size=table_config.vocabulary_size,
        embedding_dim=table_config.embedding_dim,
        initializer=keras_to_jax_initializer(initializer),
        optimizer=keras_to_jte_optimizer(table_config.optimizer),
        combiner=table_config.combiner,
        max_ids_per_partition=table_config.max_ids_per_partition,
        max_unique_ids_per_partition=table_config.max_unique_ids_per_partition,
    )


def keras_to_jte_table_configs(
    table_configs: Nested[config.TableConfig],
) -> Nested[embedding_spec.TableSpec]:
    """Converts Keras RS `TableConfig`s to JAX TPU Embedding `TableSpec`s."""
    return keras.tree.map_structure(
        _keras_to_jte_table_config,
        table_configs,
    )


def _jte_to_keras_table_config(
    table_spec: embedding_spec.TableSpec,
) -> config.TableConfig:
    return config.TableConfig(
        name=table_spec.name,
        vocabulary_size=table_spec.vocabulary_size,
        embedding_dim=table_spec.embedding_dim,
        initializer=jax_to_keras_initializer(table_spec.initializer),
        optimizer=jte_to_keras_optimizer(table_spec.optimizer),
        combiner=table_spec.combiner,
        max_ids_per_partition=table_spec.max_ids_per_partition,
        max_unique_ids_per_partition=table_spec.max_unique_ids_per_partition,
    )


def jte_to_keras_table_configs(
    table_specs: Nested[embedding_spec.TableSpec],
) -> Nested[config.TableConfig]:
    """Converts JAX TPU Embedding `TableSpec`s to Keras RS `TableConfig`s."""
    output: Nested[config.TableConfig] = keras.tree.map_structure(
        _jte_to_keras_table_config,
        table_specs,
    )
    return output


def _keras_to_jte_feature_config(
    feature_config: config.FeatureConfig,
    table_spec_map: dict[str, embedding_spec.TableSpec],
) -> embedding_spec.FeatureSpec:
    table_spec = table_spec_map.get(feature_config.table.name, None)
    if table_spec is None:
        table_spec = _keras_to_jte_table_config(feature_config.table)
        table_spec_map[feature_config.table.name] = table_spec

    return embedding_spec.FeatureSpec(
        name=feature_config.name,
        table_spec=table_spec,
        input_shape=feature_config.input_shape,
        output_shape=feature_config.output_shape,
    )


def keras_to_jte_feature_configs(
    feature_configs: Nested[config.FeatureConfig],
) -> Nested[embedding_spec.FeatureSpec]:
    """Converts Keras RS `FeatureConfig`s to JAX TPU Embedding `FeatureSpec`s.

    Args:
        feature_configs: Keras RS feature configurations.

    Returns:
        JAX TPU Embedding feature specifications.
    """
    table_spec_map: dict[str, embedding_spec.TableSpec] = {}
    return keras.tree.map_structure(
        lambda feature_config: _keras_to_jte_feature_config(
            feature_config, table_spec_map
        ),
        feature_configs,
    )


def _jte_to_keras_feature_config(
    feature_spec: embedding_spec.FeatureSpec,
    table_config_map: dict[str, config.TableConfig],
) -> config.FeatureConfig:
    table_config = table_config_map.get(feature_spec.table_spec.name, None)
    if table_config is None:
        table_config = _jte_to_keras_table_config(feature_spec.table_spec)
        table_config_map[feature_spec.table_spec.name] = table_config

    return config.FeatureConfig(
        name=feature_spec.name,
        table=table_config,
        input_shape=feature_spec.input_shape,
        output_shape=feature_spec.output_shape,
    )


def jte_to_keras_feature_configs(
    feature_specs: Nested[embedding_spec.FeatureSpec],
) -> Nested[config.FeatureConfig]:
    """Converts JAX TPU Embedding `FeatureSpec`s to Keras RS `FeatureConfig`s.

    Args:
        feature_specs: JAX TPU Embedding feature specifications.

    Returns:
        Keras RS feature configurations.
    """
    table_config_map: dict[str, config.TableConfig] = {}
    output: Nested[config.FeatureConfig] = keras.tree.map_structure(
        lambda feature_spec: _jte_to_keras_feature_config(
            feature_spec, table_config_map
        ),
        feature_specs,
    )
    return output
